from dataclasses import dataclass
from typing import Callable, Any

from menschmachine.llm.types import Agent, ToolProperty
from menschmachine.patch.xml_patch import XML_FORMAT


@dataclass
class FileContentAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], Any]):
        super().__init__(
            name="get_file_content",
            description="Get the content of a file in the repository",
            properties=[
                ToolProperty(
                    name="filepath",
                    description="The complete path of the file as presented in the file_list array",
                    type="string",
                    required=True
                )
            ],
            callback=callback,
        )


# TODO make it a persistent shell during the agent loop
"""
import code

class InteractiveShell:
    def __init__(self):
        self.variables = {}
        self.interpreter = code.InteractiveInterpreter(self.variables)

    def execute(self, command):
        return self.interpreter.runcode(command)

    def get_variable(self, var_name):
        return self.variables.get(var_name)

# Example usage
shell = InteractiveShell()

# Execute commands
shell.execute("x = 5")
shell.execute("y = 10")
shell.execute("result = x + y")
shell.execute("print(f'The result is: {result}')")
"""


@dataclass
class RunShellCommandAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], Any]):
        super().__init__(
            name="run_shell_command",
            description="Run a shell command on behalf of the user and receive the output. The command will be run like this: subprocess.run(command, shell=True, capture_output=True, text=True)",
            properties=[
                ToolProperty(
                    name="command",
                    description="The shell command to run",
                    type="string",
                    required=True
                )
            ],
            callback=callback
        )


@dataclass
class CreateFileAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], Any]):
        super().__init__(
            name="create_file",
            description="Create a new file in the repository, use only for files which do not exist yet. If the file exists, create a patch in the xml format shown to you.",
            properties=[
                ToolProperty(
                    name="filepath",
                    description="The complete path of the file to create",
                ),
                ToolProperty(
                    name="content",
                    description="The content of the new file to create",
                )
            ],
            callback=callback
        )


@dataclass
class SearchFilesAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], Any]):
        super().__init__(
            name="search_files",
            description="Perform a regex search inside the contents of all files in the repository, returning a list of all files which contents match the regex. "
                        "If you want to search for filenames, use the list_files tool.",
            properties=[
                ToolProperty(
                    name="regex",
                    description="The regular expression pattern to search for. Uses Rust regex syntax."
                )
            ],
            callback=callback
        )


@dataclass
class AskFollowUpQuestionAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], Any]):
        super().__init__(
            name="ask_followup_question",
            description="""Ask the user a question to gather additional information needed to complete the task. 
                This tool should be used when you encounter ambiguities, need clarification, or require more details to proceed effectively. 
                It allows for interactive problem-solving by enabling direct communication with the user. 
                Use this tool judiciously to maintain a balance between gathering necessary information and avoiding excessive back-and-forth.",
                """,
            properties=[
                ToolProperty(
                    name="question",
                    description="The question to ask the user. This should be a clear, specific question that addresses the information you need."
                )
            ],
            callback=callback
        )


@dataclass
class PatchAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], str]):
        super().__init__(
            name="patch_file",
            description="Change the contents of an existing file",
            properties=[
                ToolProperty(
                    name="xml_patch",
                    description=f"Provide the suggested changes strictly in the format `{XML_FORMAT}`, one by one. If you have multiple changes, call this tool multiple times. Always use get_file_content after a patch to receive the updated file contents.",
                    type="string",
                    required=True
                )
            ],
            callback=callback
        )


@dataclass
class FileListAgent(Agent):
    def __init__(self, callback: Callable[[str, dict], str]):
        super().__init__(
            name="list_files",
            description="List all files in the repository",
            properties=[],
            callback=callback
        )



from abc import ABC, abstractmethod
from typing import Callable, Any, Union, Awaitable


class Callbacks(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def get_file_list_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_question_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_run_shell_command_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_search_files_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_file_content_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_patch_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_create_file_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_delete_file_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_create_directory_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass

    @abstractmethod
    def get_delete_directory_callback(self) -> Callable[[Any], Union[str, Awaitable[str]]]:
        pass



import json
import os
from typing import List, Any

from openai import OpenAI, RateLimitError
from openai.types.chat import ChatCompletion, ChatCompletionMessage, ChatCompletionMessageToolCall

from menschmachine.event import events
from menschmachine.llm.types import FileRequestCallbackType, QuestionCallbackType, CreateFileCallbackType, \
    SearchFilesCallbackType, \
    ApiInterface, ApiResponse, MaxNumberOfLoops, ApiCosts, Agent, ModelRegistry
from menschmachine.log import get_logger
from menschmachine.util import symetric_dict

_MODEL_REGISTRY = ModelRegistry("haiku")

# Register models with their aliases
_MODEL_REGISTRY.register("anthropic/claude-3-haiku:beta", ["haiku3", "claude-haiku3", "haiku-3", "claude-haiku-3"],
                         0.25, 1.25)
_MODEL_REGISTRY.register("anthropic/claude-3.5-sonnet:beta", ["sonnet", "claude-sonnet"], 3, 15)
_MODEL_REGISTRY.register("anthropic/claude-3-7-sonnet:beta", ["sonnet-3.7", "claude-sonnet-3.7"], 3, 15)
_MODEL_REGISTRY.register("anthropic/claude-3-opus:beta", ["opus", "claude-opus"], 15, 75)
_MODEL_REGISTRY.register("openai/gpt-4o-2024-11-20", ["gpt4o", "gpt4", "gpt-4o"], 2.5, 10)
_MODEL_REGISTRY.register("openai/o1-mini-2024-09-12", ["gpt4o-mini", "gpt4-mini"], 3, 12)
_MODEL_REGISTRY.register("openai/o1", ["o1", "gpt-01"], 15, 60)
_MODEL_REGISTRY.register("openai/o3-mini", ["o3-mini", "gpt-03-mini", "gpt-03mini"], 1.1, 4.4)
_MODEL_REGISTRY.register("deepseek/deepseek-chat", ["deepseek-chat"], 0.14, 0.28)
_MODEL_REGISTRY.register("deepseek/deepseek-r1", ["deepseek", "r1"], 0.55, 2.19)
_MODEL_REGISTRY.register("google/gemini-pro-1.5", ["gemini-pro"], 1.25, 5)
_MODEL_REGISTRY.register("google/gemini-flash-1.5", ["gemini-flash"], 0.075, 0.3)
_MODEL_REGISTRY.register("openai/o1-preview-2024-09-12", ["o1", "gpt-o1", "o1-preview"], 15, 60)
_MODEL_REGISTRY.register("anthropic/claude-3-5-haiku:beta",
                         ["haiku-3.5", "claude-haiku-3.5", "claude-haiku",
                          "haiku-35", "claude-haiku-35", "haiku",
                          "haiku-3-5", "claude-haiku-3-5"],
                         1., 5.)
_MODEL_REGISTRY.register("perplexity/sonar", ["perplexity", "sonar"], 1., 1.)
_MODEL_REGISTRY.register("openai/gpt-4.5-preview", ["gpt45", "gpt-45", "gpt4.5"], 68, 150)
_MODEL_REGISTRY.register("openai/gpt-4o-mini-2024-07-18", ["gpt4o-mini", "gpt-4o-mini"], 0.15, 0.6)


class OpenRouterApi(ApiInterface):

    def __init__(self):
        super().__init__()

    def raw(self, messages: list[dict],
            tools: list[dict] = None,
            tool_choice: dict | str | None = None,
            model: str = None,
            max_tokens: int = 4000,
            temperature=0.0,
            ) -> ChatCompletion:
        client = self.create_client()
        extra_headers = self.get_extra_headers()
        model = self.get_model(model)
        if temperature is None:
            temperature = 0.0
        try:
            result = None
            if tools is None:
                result = client.chat.completions.create(
                    max_tokens=max_tokens,
                    temperature=temperature,
                    messages=messages,
                    model=model,
                    extra_headers=extra_headers
                )
            else:
                if tool_choice is None:
                    result = client.chat.completions.create(
                        max_tokens=max_tokens,
                        temperature=temperature,
                        messages=messages,
                        tools=tools,
                        model=model,
                        extra_headers=extra_headers
                    )
                else:
                    result = client.chat.completions.create(
                        max_tokens=max_tokens,
                        temperature=temperature,
                        messages=messages,
                        tools=tools,
                        tool_choice=tool_choice,
                        model=model,
                        extra_headers=extra_headers
                    )
            return result

        except RateLimitError as e:
            headers = e.response.headers
            rate_limit_info = {
                'x-ratelimit-limit-requests': headers.get('x-ratelimit-limit-requests'),
                'x-ratelimit-remaining-requests': headers.get('x-ratelimit-remaining-requests'),
                'x-ratelimit-reset-requests': headers.get('x-ratelimit-reset-requests'),
                'x-ratelimit-limit-tokens': headers.get('x-ratelimit-limit-tokens'),
                'x-ratelimit-remaining-tokens': headers.get('x-ratelimit-remaining-tokens'),
                'x-ratelimit-reset-tokens': headers.get('x-ratelimit-reset-tokens'),
                'retry-after': headers.get('retry-after'),
            }

            # Create a log message to show exceeded limits
            exceeded_limits = []

            # Check for exceeded limits
            if int(rate_limit_info['x-ratelimit-remaining-tokens']) == 0:
                exceeded_limits.append('Tokens limit exceeded')

            if int(rate_limit_info['x-ratelimit-remaining-requests']) == 0:
                exceeded_limits.append('Requests limit exceeded')

            # Prepare the log message
            if exceeded_limits:
                get_logger().error(f"Rate limit exceed: [{', '.join(exceeded_limits)}]")
                exceeded_info = {key: value for key, value in rate_limit_info.items() if value is not None}
                get_logger().debug(f"Rate limits exceeded: {exceeded_info}")
            else:
                get_logger().info("No rate limits exceeded.")
            raise e

    def get_extra_headers(self):
        extra_headers = {
            "HTTP-Referer": "https://menschmachine.com",
            "X-Title": "MenschMachine"
        }
        return extra_headers

    def create_client(self):
        client = OpenAI(
            api_key=os.environ["OPENROUTER_API_KEY"],
            base_url="https://openrouter.ai/api/v1"
        )
        return client

    @staticmethod
    def token_counts_from_response(response):
        usage = response.usage
        if usage is not None:
            prompt_tokens = usage.prompt_tokens
            completion_tokens = usage.completion_tokens
        else:
            prompt_tokens = -1
            completion_tokens = -1
        return completion_tokens, prompt_tokens

    def is_model_supported(self, model: str) -> bool:
        return self.get_model_registry().is_supported(model)

    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 10) -> ApiResponse:

        costs = self.get_api_costs(model, 0, 0)
        agent_map = self.get_agent_map(agent_ids)
        tools = self.tools_from_agent_map(agent_map)
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        events.emit(f"/mm/agent/init/{uuid}", uuid, "openrouter", symetric_dict(prompt, model, tools, messages))
        return await self.loop(uuid, model, agent_map, 0, max_nr_of_loops, messages, temperature, costs)

    def tools_from_agent_map(self, agent_map):
        tools = []
        for agent_id in agent_map:
            agent = agent_map[agent_id]
            properties, required_properties = self.properties_for_agent(agent)
            input_schema = self.input_schema_from_properties(properties, required_properties)
            tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": agent.name,
                        "description": agent.description,
                        "parameters": input_schema
                    }
                }
            )
        return tools

    def ping(self) -> ApiResponse:
        return self.ask("ping", self.get_long_name("haiku"))

    def ask(self,
            prompt: str,
            model: str = None,
            max_tokens: int = 4096,
            temperature=0.0,
            assistant_msg=None) -> ApiResponse:

        messages = [{"role": "user", "content": prompt}]
        if assistant_msg:
            messages.append({"role": "assistant", "content": assistant_msg})

        model = self.get_model(model)
        response = self.raw(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature
        )
        completion_tokens, prompt_tokens = self.token_counts_from_response(response)
        costs = self.get_api_costs(model, prompt_tokens, completion_tokens)
        answer = response.choices[0].message.content

        return ApiResponse(message=answer, error=None, costs=costs)

    async def with_tools_loop(self, prompt: str,
                              file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType,
                              model: str,
                              max_tokens: int = 100000,
                              temperature: float = 0.1) -> ApiResponse:
        model = self.get_model(model)
        tools = [
            {
                "type": "function",
                "function": {
                    "name": "get_file_content",
                    "description": "Get the content of a file in the repository",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "filepath": {
                                "type": "string",
                                "description": "The complete path of the file as presented in the file_list array"
                            }
                        },
                        "required": ["filepath"]
                    }
                }
            },
            {
                "type": "function",
                "function": {
                    "name": "create_file",
                    "description": "Create a new file in the repository, use only for files which do not exist yet. If the file exists, create a patch in the xml format shown to you.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "filepath": {
                                "type": "string",
                                "description": "The complete path of the file to create"
                            },
                            "content": {
                                "type": "string",
                                "description": "The content of the new file to create"
                            }
                        },
                        "required": ["filepath", "content"]
                    }
                }
            },
            {
                "type": "function",
                "function": {
                    "name": "search_files",
                    "description": "Perform a regex search across all files in the repository, returning a list of all filenames matching the regex.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                                "description": "The regular expression pattern to search for. Uses Rust regex syntax.",
                            }
                        },
                        "required": ["regex"],
                    }
                }
            },
            {
                "type": "function",
                "function": {
                    "name": "ask_followup_question",
                    "description": "Ask the user a question to gather additional information needed to complete the task.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "question": {
                                "type": "string",
                                "description": "The question to ask the user. This should be a clear, specific question that addresses the information you need.",
                            },
                        },
                        "required": ["question"],
                    }
                }
            },
        ]

        messages: list[dict[str, str] | ChatCompletionMessage] = [{"role": "user", "content": prompt}]
        files = {}
        costs = self.get_api_costs(model, 0, 0)

        async def loop() -> ApiResponse:

            response = self.raw(
                model=model,
                messages=messages,
                tools=tools,
                tool_choice="auto",
                max_tokens=max_tokens,
                temperature=temperature,
            )

            message = response.choices[0].message
            messages.append(message)
            completion_tokens, prompt_tokens = self.token_counts_from_response(response)
            costs.add(self.get_api_costs(model, prompt_tokens, completion_tokens))

            if message.tool_calls is not None and len(message.tool_calls) > 0:
                tool_calls: List[ChatCompletionMessageToolCall] = message.tool_calls
                tool_call_results = list()
                for tool_call in tool_calls:
                    function_name = tool_call.function.name
                    function_args = tool_call.function.arguments
                    tool_call_id = tool_call.id

                    if function_name == "get_file_content":
                        tool_call_results.append(
                            {"role": "tool", "tool_call_id": tool_call_id,
                             "content": await get_file_content_exec(function_args)})
                    elif function_name == "ask_followup_question":
                        tool_call_results.append(
                            {"role": "tool", "tool_call_id": tool_call_id,
                             "content": await ask_followup_question(function_args)})
                    elif function_name == "search_files":
                        tool_call_results.append(
                            {"role": "tool", "tool_call_id": tool_call_id,
                             "content": await search_files(function_args)})
                    elif function_name == "create_file":
                        tool_call_results.append(
                            {"role": "tool", "tool_call_id": tool_call_id, "content": await create_file(function_args)})

                messages.extend(tool_call_results)
                return await loop()
            else:
                return ApiResponse(message=message.content, error=None, costs=costs, files=files)

        async def create_file(args):
            args_dict = json.loads(args)
            filepath = args_dict['filepath']
            content = args_dict['content']
            return await create_file_callback(filepath, content)

        async def search_files(args):
            args_dict = json.loads(args)
            q = args_dict['regex']
            files_ = await search_files_callback(q)
            return json.dumps(files_)

        async def ask_followup_question(args):
            args_dict = json.loads(args)
            q = args_dict['question']
            a = await follow_up_question_callback(q)
            return a

        async def get_file_content_exec(args):
            args_dict = json.loads(args)
            files_requested = [args_dict['filepath']]
            file_content_result = await file_content_callback(files_requested)
            for file in file_content_result:
                files[file] = file_content_result[file]
            return json.dumps(file_content_result)

        return await loop()

    async def loop(self,
                   uuid: str,
                   model: str,
                   agent_map: dict[str, Agent],
                   nr_of_loops: int,
                   max_nr_of_loops: int,
                   messages: list,
                   temperature: float,
                   costs: ApiCosts) -> Any:

        async def append_next_step(tool_result: Any, _response: str, function_name: str, _tool_use_id: str):
            messages.append({"role": "assistant", "content": _response})
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": _tool_use_id,
                    "name": function_name,
                    "content": json.dumps(tool_result)
                })
            events.emit(f"/mm/agent/step/{uuid}", uuid, "openrouter", messages)

        nr_of_loops += 1
        get_logger().debug(f"Loop number: {nr_of_loops}")
        if nr_of_loops > max_nr_of_loops:
            get_logger().warning("Max number of loops reached")
            raise MaxNumberOfLoops()
        tools = self.tools_from_agent_map(agent_map)
        try:
            response = self.raw(
                model=model,
                messages=messages,
                tools=tools,
                temperature=temperature,
                max_tokens=4096
            )

            completion_tokens, prompt_tokens = self.token_counts_from_response(response)
            costs.add(self.get_api_costs(model, prompt_tokens, completion_tokens))
            choice = response.choices[0]
            message = choice.message
            get_logger().debug(f"Choice: {choice}")
            if choice.finish_reason == 'tool_calls':

                tool_calls: List[ChatCompletionMessageToolCall] = message.tool_calls
                tool_call_results = list[dict]()
                for tool_call in tool_calls:
                    function_args = tool_call.function.arguments
                    tool_use_id = tool_call.id
                    tool_name = tool_call.function.name
                    get_logger().info(f"Tool requested: {tool_name}")
                    get_logger().debug(f"Tool request messsage: {message}")
                    if tool_name not in agent_map:
                        raise ValueError(f"Unknown tool requested: {tool_name}")
                    # noinspection DuplicatedCode
                    try:
                        await agent_map[tool_name].call(uuid,
                                                        json.loads(function_args),
                                                        response,
                                                        tool_use_id,
                                                        append_next_step)
                    except Exception as e:
                        self.handle_tool_call_exception(message, tool_name, e)
                    await self.loop(
                        uuid,
                        model,
                        agent_map,
                        nr_of_loops,
                        max_nr_of_loops,
                        messages,
                        temperature,
                        costs)
                messages.extend(tool_call_results)
                return await self.loop(
                    uuid,
                    model,
                    agent_map,
                    nr_of_loops,
                    max_nr_of_loops,
                    messages,
                    temperature,
                    costs)
            elif choice.finish_reason == 'end_turn' or choice.finish_reason == 'STOP':
                get_logger().info("Agent loop completed successfully")
                events.emit(f"/mm/agent/completed/{uuid}", uuid, "openrouter", symetric_dict(response, costs))
                return ApiResponse(message=message.content, error=None, costs=costs)
            else:
                get_logger().warning(f"Unexpected stop reason: {choice.finish_reason}")
                return ApiResponse(message=message.content, error=None, costs=costs)
        except Exception as e:
            events.emit(f"/mm/agent/error/{uuid}", uuid, "openrouter", symetric_dict(e, costs))
            get_logger().error(f"Error in agent loop: {str(e)}")
            raise

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        def costs(nr_tokens, price_per_million: float):
            return (nr_tokens * (price_per_million / (1000. * 1000.))) / 100.  # in cent

        in_price, out_price = self.get_prices(model)
        return costs(prompt_tokens, in_price) + costs(completion_tokens, out_price)

    def get_model(self, model):
        return self.get_model_registry().get_model(model)

    def get_model_registry(self) -> ModelRegistry:
        return _MODEL_REGISTRY

    def get_prices(self, model):
        return self.get_model_registry().get_prices(model)

    def get_long_name(self, param):
        return self.get_model_registry().get_long_name(param)



import os
from typing import Any

from anthropic import Anthropic, RateLimitError
from anthropic.types import Message

from menschmachine.event import events
from menschmachine.llm.types import FileRequestCallbackType, QuestionCallbackType, CreateFileCallbackType, \
    SearchFilesCallbackType, \
    ApiInterface, ApiResponse, MaxNumberOfLoops, ApiCosts, Agent
from menschmachine.log import get_logger
from menschmachine.patch import Patch
from menschmachine.util import symetric_dict

models = {
    "claude-3-haiku-20240307": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
    },
    "claude-3-5-sonnet-20241022": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
    },
    "claude-3-5-haiku-20241022": {
        "max_input_tokens": 200000,
        "max_output_tokens": 4096,
    },
    "claude-3-7-sonnet-20250219": {
        "max_input_tokens": 200000,
        "max_output_tokens": 8192,
    }
}


def get_max_output_tokens(model: str) -> int:
    if model in models:
        return models[model]["max_output_tokens"]
    else:
        return 4096


_LATEST_HAIKU_MODEL = "claude-3-haiku-20240307"
_LATEST_HAIKU_3_5_MODEL = "claude-3-5-haiku-20241022"
_LATEST_SONNET_3_7_MODEL = "claude-3-7-sonnet-20250219"
_LATEST_SONNET_MODEL = _LATEST_SONNET_3_7_MODEL
_LATEST_OPUS_MODEL = "claude-3-opus-20240229"


def raw(messages: list[dict],
        tools: list[dict] = None,
        tool_choice: dict = None,
        model: str = _LATEST_HAIKU_MODEL,
        temperature=0.0,
        ) -> Message:
    client = Anthropic()
    model = get_model(model)
    if temperature is None:
        temperature = 0.0
    if tools is None or len(tools) == 0:
        return client.messages.create(
            max_tokens=get_max_output_tokens(model),
            temperature=temperature,
            messages=messages,
            model=model,
        )
    else:
        if tool_choice is None:
            return client.messages.create(
                max_tokens=get_max_output_tokens(model),
                temperature=temperature,
                messages=messages,
                tools=tools,
                model=model,
            )
        else:
            return client.messages.create(
                max_tokens=get_max_output_tokens(model),
                temperature=temperature,
                messages=messages,
                tools=tools,
                tool_choice=tool_choice,
                model=model,
            )


def get_model(model: str) -> str:
    if "UNIVERSAL_MODEL" in os.environ:
        model = os.environ["UNIVERSAL_MODEL"]
    if model is None:
        return _LATEST_HAIKU_MODEL
    if _get_supported_model(model) is not None:
        return _get_supported_model(model)
    return model


def _get_supported_model(model: str) -> str | None:
    if model is not None:
        if model.lower() == "haiku":
            return _LATEST_HAIKU_MODEL
        elif model.lower() == "haiku3":
            return _LATEST_HAIKU_MODEL
        elif model.lower() == "haiku-3.5":
            return _LATEST_HAIKU_3_5_MODEL
        elif model.lower() == "sonnet":
            return _LATEST_SONNET_MODEL
        elif model.lower() == "sonnet-3.7":
            return _LATEST_SONNET_3_7_MODEL
        elif model.lower() == "opus":
            return _LATEST_OPUS_MODEL
    return None


class ClaudeApi(ApiInterface):

    def __init__(self):
        super().__init__()

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        def costs(nr_tokens, price_per_million: float):
            return (nr_tokens * (price_per_million / (1000. * 1000.))) / 100.  # in cent

        if "sonnet" in model:
            return costs(prompt_tokens, 3) + costs(completion_tokens, 15)
        elif "haiku-3.5" in model:
            return costs(prompt_tokens, 1) + costs(completion_tokens, 5)
        elif "haiku" in model:
            return costs(prompt_tokens, 0.25) + costs(completion_tokens, 1.25)
        elif "opus" in model:
            return costs(prompt_tokens, 15) + costs(completion_tokens, 75)
        else:
            return -1.

    def is_model_supported(self, model: str) -> bool:
        return _get_supported_model(model) is not None

    def ping(self) -> ApiResponse:
        return self.ask("ping", _LATEST_HAIKU_MODEL)

    async def with_tools_loop(self,
                              prompt: str,
                              file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType,
                              model: str = _LATEST_HAIKU_MODEL,
                              max_tokens: int = models[_LATEST_HAIKU_MODEL]["max_output_tokens"],
                              temperature: float = 0.1,
                              ) -> ApiResponse:
        tools = [
            {
                "name": "get_file_content",
                "description": "Get the content of a file in the repository",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "filepath": {
                            "type": "string",
                            "description": "The complete path of the file as presented in the file_list array"
                        }
                    },
                    "required": ["filepath"]
                }
            },
            {
                "name": "create_file",
                "description": "Create a new file in the repository, use only for files which do not exist yet. If the file exists, create a patch in the xml format shown to you.",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "filepath": {
                            "type": "string",
                            "description": "The complete path of the file to create"
                        },
                        "content": {
                            "type": "string",
                            "description": "The content of the new file to create"
                        }
                    },
                    "required": ["filepath", "content"]
                }
            },
            {
                "name": "search_files",
                "description":
                    "Perform a regex search across all files in the repository, returning a list of all filenames matching the regex.",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "regex": {
                            "type": "string",
                            "description": "The regular expression pattern to search for. Uses Rust regex syntax.",
                        }
                    },
                    "required": ["regex"],
                },
            },
            {
                "name": "ask_followup_question",
                "description":
                    """Ask the user a question to gather additional information needed to complete the task. 
                    This tool should be used when you encounter ambiguities, need clarification, or require more details to proceed effectively. 
                    It allows for interactive problem-solving by enabling direct communication with the user. 
                    Use this tool judiciously to maintain a balance between gathering necessary information and avoiding excessive back-and-forth.",
                    """,
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "question": {
                            "type": "string",
                            "description":
                                "The question to ask the user. This should be a clear, specific question that addresses the information you need.",
                        },
                    },
                    "required": ["question"],
                },
            },
        ]
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        files = {}
        costs = self.get_api_costs(model, 0, 0)

        async def loop() -> ApiResponse:
            response = raw(messages=messages, tools=tools, model=model, temperature=temperature)
            costs.add(self.get_api_costs(model, response.usage.input_tokens, response.usage.output_tokens))
            if response.stop_reason == 'tool_use':
                msg = response.content[len(response.content) - 1]
                tool_use_id = msg.id
                tool_name = msg.name
                get_logger().info(f"Tool requested: {tool_name}")
                if tool_name == "get_file_content":
                    await get_file_content_exec(msg, response, tool_use_id)
                elif tool_name == "ask_followup_question":
                    await ask_followup_question(msg, response, tool_use_id)
                elif tool_name == "search_files":
                    await search_files(msg, response, tool_use_id)
                elif tool_name == "create_file":
                    await create_file(msg, response, tool_use_id)
                return await loop()
            elif response.stop_reason == 'max_tokens':
                messages.append({"role": "assistant", "content": response.content[0].text.strip()})
                tools.clear()  # should be last call and claude refuses to continue with tools enabled
                return await loop()
            else:
                return ApiResponse(message=response.content[0].text, error=None, costs=costs, files=files)

        async def create_file(msg, response, tool_use_id):
            filepath = msg.input['filepath']
            content = msg.input['content']
            try:
                # check if it is a patch
                patch = Patch.from_dirty_xml(content)
                if patch is not None:
                    return content
            except:
                pass
            result = await create_file_callback(filepath, content)
            await append_next_step(result, response, tool_use_id)

        async def search_files(msg, response, tool_use_id):
            q = msg.input['regex']
            files = await search_files_callback(q)
            content = []
            for f in files:
                content.append({"type": "text", "text": f})
            await append_next_step(content, response, tool_use_id)

        async def ask_followup_question(msg, response, tool_use_id):
            q = msg.input['question']
            a = await follow_up_question_callback(q)
            await append_next_step(a, response, tool_use_id)

        async def get_file_content_exec(msg, response, tool_use_id):
            files_requested = [msg.input['filepath']]
            file_content_result = await file_content_callback(files_requested)
            file_content = list()
            for file in file_content_result:
                file_content.append({"type": "text", "text": file + ":\n" + file_content_result[file]})
                files[file] = file_content_result[file]
            await append_next_step(file_content, response, tool_use_id)

        async def append_next_step(tool_result: Any, response, tool_use_id):
            messages.append({"role": "assistant", "content": response.content})
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "tool_result",
                            "tool_use_id": tool_use_id,
                            "content": tool_result
                        }
                    ]
                })

        return await loop()

    def ask(self,
            prompt: str,
            model: str = _LATEST_HAIKU_MODEL,
            max_tokens: int = models[_LATEST_HAIKU_MODEL]["max_output_tokens"],
            temperature=0.0,
            assistant_msg=None
            ) -> ApiResponse:
        model = get_model(model)

        client = Anthropic()
        if temperature is None:
            temperature = 0.0
        try:
            if assistant_msg is None:
                message = client.messages.create(
                    temperature=temperature,
                    max_tokens=max_tokens,
                    messages=[
                        {
                            "role": "user",
                            "content": prompt,
                        }
                    ],
                    model=model,
                )
            else:
                message = client.messages.create(
                    temperature=temperature,
                    max_tokens=max_tokens,
                    messages=[
                        {
                            "role": "user",
                            "content": prompt,
                        },
                        {
                            "role": "assistant",
                            "content": assistant_msg
                        }
                    ],
                    model=model,
                )
            answer = message.content[0].text
            costs = self.get_api_costs(model, message.usage.input_tokens, message.usage.output_tokens)

            return ApiResponse(message=answer, error=None, costs=costs)

        except RateLimitError as e:
            headers = e.response.headers
            rate_limit_info = {
                'x-should-retry': headers.get('x-should-retry'),
                'anthropic-ratelimit-requests-limit': headers.get('anthropic-ratelimit-requests-limit'),
                'anthropic-ratelimit-requests-remaining': headers.get('anthropic-ratelimit-requests-remaining'),
                'anthropic-ratelimit-requests-reset': headers.get('anthropic-ratelimit-requests-reset'),
                'anthropic-ratelimit-tokens-limit': headers.get('anthropic-ratelimit-tokens-limit'),
                'anthropic-ratelimit-tokens-remaining': headers.get('anthropic-ratelimit-tokens-remaining'),
                'anthropic-ratelimit-tokens-reset': headers.get('anthropic-ratelimit-tokens-reset'),
                'retry-after': headers.get('retry-after'),
            }

            # Create a log message to show exceeded limits
            exceeded_limits = []

            # Check for exceeded limits
            if int(rate_limit_info['anthropic-ratelimit-tokens-remaining']) == 0:
                exceeded_limits.append('Tokens limit exceeded')

            if int(rate_limit_info['anthropic-ratelimit-requests-remaining']) == 0:
                exceeded_limits.append('Requests limit exceeded')

            # Prepare the log message
            if exceeded_limits:
                get_logger().error(f"Rate limit exceed: [{', '.join(exceeded_limits)}]")
                exceeded_info = {key: value for key, value in rate_limit_info.items() if value is not None}
                get_logger().debug(f"Rate limits exceeded: {exceeded_info}")
            else:
                get_logger().info("No rate limits exceeded.")
            raise e

    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str = _LATEST_HAIKU_MODEL,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 100
                         ) -> ApiResponse:

        model = get_model(model)
        get_logger().debug(f"Using model: {model}")
        agent_map = self.get_agent_map(agent_ids)
        tools = self.tools_from_agent_map(agent_map)
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        events.emit(f"/mm/agent/init/{uuid}", uuid, "claude",
                    symetric_dict(prompt, model, tools, messages, agent_ids, max_nr_of_loops, temperature))

        costs = self.get_api_costs(model, 0, 0)
        return await self.loop(uuid, model, agent_map, 0, max_nr_of_loops, messages, temperature, costs)

    def tools_from_agent_map(self, agent_map):
        tools = []
        for agent_name in agent_map:
            agent = agent_map[agent_name]
            properties, required_properties = self.properties_for_agent(agent)
            input_schema = self.input_schema_from_properties(properties, required_properties)
            tools.append(
                {
                    "name": agent.name,
                    "description": agent.description,
                    "input_schema": input_schema
                }
            )
        return tools

    async def loop(self,
                   uuid: str,
                   model: str,
                   agent_map: dict[str, Agent],
                   nr_of_loops: int,
                   max_nr_of_loops: int,
                   messages: list,
                   temperature: float,
                   costs: ApiCosts) -> ApiResponse:

        async def append_next_step(tool_result: Any, _response: Message, function_name: str, _tool_use_id: str):
            get_logger().debug(f"append_next_step: {tool_result}")
            messages.append({"role": "assistant", "content": _response.content})
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "tool_result",
                            "tool_use_id": _tool_use_id,
                            "content": tool_result
                        }
                    ]
                })
            events.emit(f"/mm/agent/step/{uuid}", uuid, "claude", messages)

        nr_of_loops += 1
        get_logger().debug(f"loop: {nr_of_loops}")
        if nr_of_loops > max_nr_of_loops:
            get_logger().warning("Max number of loops reached")
            raise MaxNumberOfLoops()
        try:
            tools = self.tools_from_agent_map(agent_map)
            response = raw(messages=messages, tools=tools, model=model, temperature=temperature)
            get_logger().debug(f"response: {response} in loop: {nr_of_loops}")
            costs.add(self.get_api_costs(model, response.usage.input_tokens, response.usage.output_tokens))
            msg = response.content[len(response.content) - 1]
            if response.stop_reason == 'tool_use':
                tool_use_id = msg.id
                tool_name = msg.name
                get_logger().info(f"Tool requested: {tool_name}")
                get_logger().debug(f"Tool request messsage: {msg}")
                if tool_name not in agent_map:
                    raise ValueError(f"Unknown tool requested: {tool_name}")
                try:
                    await agent_map[tool_name].call(uuid, msg.input, response, tool_use_id, append_next_step)
                except Exception as e:
                    self.handle_tool_call_exception(messages, tool_name, e)
                return await self.loop(uuid,
                                       model,
                                       agent_map,
                                       nr_of_loops,
                                       max_nr_of_loops,
                                       messages,
                                       temperature,
                                       costs)
            elif response.stop_reason == 'max_tokens':
                messages.append({"role": "assistant", "content": response.content[0].text.strip()})
                return await self.loop(uuid,
                                       model,
                                       dict(),  # should be last call and claude refuses to continue with tools enabled
                                       nr_of_loops,
                                       max_nr_of_loops,
                                       messages,
                                       temperature,
                                       costs)
            elif response.stop_reason == 'end_turn':
                get_logger().info("Agent loop completed successfully")
                events.emit(f"/mm/agent/completed/{uuid}", uuid, "claude", symetric_dict(response, costs))
                return ApiResponse(message=msg.text, error=None, costs=costs)
            else:
                get_logger().warning(f"Unexpected stop reason: {response.stop_reason}")
                return ApiResponse(message=msg.text, error=None, costs=costs)
        except Exception as e:
            events.emit(f"/mm/agent/error/{uuid}", uuid, "claude", symetric_dict(e, costs))
            get_logger().error(f"Error in agent loop: {str(e)}")
            raise



import json
import os
import uuid as uuidgen
from typing import Any

import google.generativeai as genai
from google.api_core import exceptions
from google.generativeai.types import GenerateContentResponse

from menschmachine.event import events
from menschmachine.llm.types import FileRequestCallbackType, QuestionCallbackType, CreateFileCallbackType, \
    SearchFilesCallbackType, \
    ApiInterface, ApiResponse, MaxNumberOfLoops, ApiCosts, Agent, ModelRegistry
from menschmachine.log import get_logger
from menschmachine.patch import Patch
from menschmachine.util import symetric_dict

MODEL_REGISTRY = ModelRegistry("gemini-pro")

# Register models with their aliases
MODEL_REGISTRY.register("gemini-pro", ["gemini-pro"], 0.25, 0.5, "gemini-pro")
MODEL_REGISTRY.register("gemini-pro-vision", ["gemini-pro-vision"], 0.25, 0.5, "gemini-pro-vision")
MODEL_REGISTRY.register("gemini-ultra", ["gemini-ultra"], 1.25, 2.5, "gemini-ultra")
MODEL_REGISTRY.register("gemini-flash", ["gemini-flash"], 0.075, 0.3, "gemini-1.5-flash-latest")


def raw(messages: list[dict],
        tools: list[dict] = None,
        tool_choice: dict | str | None = None,
        model: str = None,
        max_tokens: int = 4000,
        temperature=0.0,
        ) -> GenerateContentResponse:
    client = create_client()
    model = MODEL_REGISTRY.get_model(model)
    official_name = MODEL_REGISTRY.get_official_name(model)

    if temperature is None:
        temperature = 0.0
    try:
        chat = client.GenerativeModel(official_name).start_chat()
        if tools is None or len(tools) == 0:
            result = chat.send_message(
                content=messages[-1]['content'],
                generation_config=genai.types.GenerationConfig(
                    max_output_tokens=max_tokens,
                    temperature=temperature,
                )
            )
        else:
            result = chat.send_message(
                content=messages[-1]['content'],
                generation_config=genai.types.GenerationConfig(
                    max_output_tokens=max_tokens,
                    temperature=temperature,
                ),
                tools=tools
            )
        return result

    except exceptions.GoogleAPICallError as e:
        get_logger().error(f"Google API Call Error: {str(e)}")
        raise e


def create_client():
    genai.configure(api_key=os.environ["GEMINI_API_KEY"])
    return genai


class GeminiApi(ApiInterface):

    def __init__(self):
        super().__init__()

    def is_model_supported(self, model: str) -> bool:
        return MODEL_REGISTRY.is_supported(model)

    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 10) -> ApiResponse:

        costs = self.get_api_costs(model, 0, 0)
        agent_map = self.get_agent_map(agent_ids)
        tools = self.tools_from_agent_map(agent_map)
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        events.emit(f"/mm/agent/init/{uuid}", uuid, "google_ai", symetric_dict(prompt, model, tools, messages))
        return await self.loop(uuid, model, agent_map, 0, max_nr_of_loops, messages, temperature, costs)

    def tools_from_agent_map(self, agent_map):
        tools = []
        for agent_id in agent_map:
            agent = agent_map[agent_id]
            properties, required_properties = self.properties_for_agent(agent)
            input_schema = self.input_schema_from_properties(properties, required_properties)
            tools.append(
                {
                    "name": agent.name,
                    "description": agent.description,
                    "parameters": input_schema
                },
            )
        return tools

    def ping(self) -> ApiResponse:
        return self.ask("ping", MODEL_REGISTRY.get_long_name("gemini-pro"))

    def ask(self,
            prompt: str,
            model: str = None,
            max_tokens: int = 4096,
            temperature=0.0,
            assistant_msg=None) -> ApiResponse:

        messages = [{"role": "user", "content": prompt}]
        if assistant_msg:
            messages.append({"role": "assistant", "content": assistant_msg})

        model = MODEL_REGISTRY.get_model(model)
        response = raw(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature
        )
        costs = self.get_api_costs(model, response.usage_metadata.prompt_token_count,
                                   response.usage_metadata.candidates_token_count)
        answer = response.text
        return ApiResponse(message=answer, error=None, costs=costs)

    async def with_tools_loop(self,
                              prompt: str,
                              file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType,
                              model: str,
                              max_tokens: int = 4000,
                              temperature: float = 0.1,
                              ) -> ApiResponse:
        tools = [
            {
                "function_declarations": [
                    {
                        "name": "get_file_content",
                        "description": "Get the content of a file in the repository",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "filepath": {
                                    "type": "string",
                                    "description": "The complete path of the file as presented in the file_list array"
                                }
                            },
                            "required": ["filepath"]
                        }
                    },
                    {
                        "name": "create_file",
                        "description": "Create a new file in the repository, use only for files which do not exist yet. If the file exists, create a patch in the xml format shown to you.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "filepath": {
                                    "type": "string",
                                    "description": "The complete path of the file to create"
                                },
                                "content": {
                                    "type": "string",
                                    "description": "The content of the new file to create"
                                }
                            },
                            "required": ["filepath", "content"]
                        }
                    },
                    {
                        "name": "search_files",
                        "description": "Perform a regex search across all files in the repository, returning a list of all filenames matching the regex.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "regex": {
                                    "type": "string",
                                    "description": "The regular expression pattern to search for. Uses Rust regex syntax.",
                                }
                            },
                            "required": ["regex"],
                        }
                    },
                    {
                        "name": "ask_followup_question",
                        "description": "Ask the user a question to gather additional information needed to complete the task.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "question": {
                                    "type": "string",
                                    "description": "The question to ask the user. This should be a clear, specific question that addresses the information you need.",
                                },
                            },
                            "required": ["question"],
                        }
                    }
                ]
            }
        ]
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        files = {}
        costs = self.get_api_costs(model, 0, 0)

        async def loop() -> ApiResponse:
            response = raw(messages=messages, tools=tools, model=model, temperature=temperature)
            costs.add(self.get_api_costs(model, response.usage_metadata.prompt_token_count,
                                         response.usage_metadata.candidates_token_count))

            if response.candidates[0].content.parts[-1].function_call:
                function_call = response.candidates[0].content.parts[-1].function_call
                tool_name = function_call.name
                get_logger().info(f"Tool requested: {tool_name}")
                if tool_name == "get_file_content":
                    await get_file_content_exec(function_call)
                elif tool_name == "ask_followup_question":
                    await ask_followup_question(function_call)
                elif tool_name == "search_files":
                    await search_files(function_call)
                elif tool_name == "create_file":
                    await create_file(function_call)
                return await loop()
            else:
                return ApiResponse(message=response.text, error=None, costs=costs, files=files)

        async def create_file(function_call):
            args = json.loads(function_call.args)
            filepath = args['filepath']
            content = args['content']
            try:
                patch = Patch.from_dirty_xml(content)
                if patch is not None:
                    return content
            except:
                pass
            result = await create_file_callback(filepath, content)
            await append_next_step(result, function_call)

        async def search_files(function_call):
            args = json.loads(function_call.args)
            q = args['regex']
            files = await search_files_callback(q)
            await append_next_step(files, function_call)

        async def ask_followup_question(function_call):
            args = json.loads(function_call.args)
            q = args['question']
            a = await follow_up_question_callback(q)
            await append_next_step(a, function_call)

        async def get_file_content_exec(function_call):
            args = json.loads(function_call.args)
            files_requested = [args['filepath']]
            file_content_result = await file_content_callback(files_requested)
            for file in file_content_result:
                files[file] = file_content_result[file]
            await append_next_step(file_content_result, function_call)

        async def append_next_step(tool_result: Any, function_call):
            messages.append({"role": "model", "content": json.dumps(function_call.dict())})
            messages.append(
                {
                    "role": "function",
                    "name": function_call.name,
                    "content": json.dumps(tool_result)
                })

        return await loop()

    async def loop(self,
                   uuid: str,
                   model: str,
                   agent_map: dict[str, Agent],
                   nr_of_loops: int,
                   max_nr_of_loops: int,
                   messages: list,
                   temperature: float,
                   costs: ApiCosts) -> Any:

        async def append_next_step(tool_result: Any, _response: str, function_name: str, tool_call_id: str):
            messages.append({"role": "assistant", "content": _response})
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call_id,
                    "name": function_name,
                    "content": json.dumps(tool_result),
                })
            events.emit(f"/mm/agent/step/{uuid}", uuid, "google_ai", messages)

        nr_of_loops += 1
        get_logger().debug(f"Loop number: {nr_of_loops}")
        if nr_of_loops > max_nr_of_loops:
            get_logger().warning("Max number of loops reached")
            raise MaxNumberOfLoops()
        tools = [
            {
                "function_declarations": self.tools_from_agent_map(agent_map)
            }
        ]
        try:
            response = raw(
                model=model,
                messages=messages,
                tools=tools,
                temperature=temperature,
                max_tokens=4096
            )
            costs.add(self.get_api_costs(model, response.usage_metadata.prompt_token_count,
                                         response.usage_metadata.candidates_token_count))
            message = response.text
            get_logger().debug(f"Response: {message}")

            # Google AI doesn't have a direct equivalent to OpenAI's tool_calls
            # We need to parse the response to determine if a tool was called
            tool_call_results = []
            for tool in tools:
                tool_name = tool['function']['name']
                if tool_name in message:
                    # Extract arguments from the message (this is a simplification)
                    function_args = json.dumps({})  # You'll need to implement proper argument extraction
                    tool_use_id = str(uuidgen.uuid4())  # Generate a unique ID for the tool call
                    get_logger().info(f"Tool requested: {tool_name}")
                    get_logger().debug(f"Tool request message: {message}")
                    if tool_name not in agent_map:
                        raise ValueError(f"Unknown tool requested: {tool_name}")
                    try:
                        await agent_map[tool_name].call(uuid,
                                                        json.loads(function_args),
                                                        response,
                                                        tool_use_id,
                                                        append_next_step)
                    except Exception as e:
                        self.handle_tool_call_exception(message, tool_name, e)
                    tool_call_results.append({
                        "role": "tool",
                        "tool_call_id": tool_use_id,
                        "name": tool_name,
                        "content": json.dumps(function_args),
                    })

            if tool_call_results:
                messages.extend(tool_call_results)
                return await self.loop(
                    uuid,
                    model,
                    agent_map,
                    nr_of_loops,
                    max_nr_of_loops,
                    messages,
                    temperature,
                    costs)
            else:
                get_logger().info("Agent loop completed successfully")
                events.emit(f"/mm/agent/completed/{uuid}", uuid, "google_ai", symetric_dict(response, costs))
                return ApiResponse(message=message, error=None, costs=costs)
        except Exception as e:
            events.emit(f"/mm/agent/error/{uuid}", uuid, "google_ai", symetric_dict(e, costs))
            get_logger().error(f"Error in agent loop: {str(e)}")
            raise

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        def costs(nr_tokens, price_per_million: float):
            return (nr_tokens * (price_per_million / (1000. * 1000.))) / 100.  # in cent

        in_price, out_price = MODEL_REGISTRY.get_prices(model)
        return costs(prompt_tokens, in_price) + costs(completion_tokens, out_price)



import logging
import os
import sys
import tempfile

import git
from git import Repo, Tree


def get_name_from_url(url: str):
    path = url.split("/")[-1]
    if path.endswith(".git"):
        return path[0:-4]
    else:
        return path


def _repo_from_url(url):
    return {
        'owner': None,
        'repository': get_name_from_url(url),
        'url': url
    }


class GitClient(object):

    def __init__(self, directory: str = ".", logger: logging.Logger = None):
        super().__init__()
        if logger is None:
            logger = logging.getLogger("menschmachine.gitclient")
        try:
            root = self._find_git_root(directory)
            self.root_dir = root
            os.chdir(root)
            self.repo = Repo(root)
            self.repo_url = self._get_repo()["url"]
        except git.exc.InvalidGitRepositoryError:
            logger.error("Current directory is not a valid git repository")
            sys.exit(1)

    def _find_git_root(self, directory):
        """
        Recursively search for the git root directory.
        """
        if os.path.isdir(os.path.join(directory, ".git")):
            return directory
        parent_dir = os.path.abspath(os.path.join(directory, ".."))
        if parent_dir == directory:
            return None
        return self._find_git_root(parent_dir)

    def _get_repo(self):
        for remote in self.repo.remotes:
            for url in remote.urls:
                return _repo_from_url(url)
        return _repo_from_url(f"file:///{self.repo.common_dir}")

    def get_diff(self):
        t = self.repo.head.commit.tree
        return self.repo.git.diff(t)

    def get_untracked_files(self):
        return self.repo.untracked_files

    def get_git_files(self):
        files = list()
        tree = self.repo.tree()
        self._read_tree(files, tree)
        return files

    def _read_tree(self, files: list[str], tree: Tree):
        for f in tree:
            if f.type == "blob":
                files.append(f.path)
            elif f.type == "tree":
                self._read_tree(files, f)

    def commit(self, message, add_all: bool = False):
        if add_all:
            self.repo.git.add(A=True)
            self.repo.git.refresh()
        index = self.repo.index
        index.commit(message)

    def diff_file(self, filename: str, content: str) -> str:
        with tempfile.NamedTemporaryFile(mode='w+', delete=True) as tmp:
            tmp.write(content)
            tmp.flush()
            return self.repo.git.diff("--no-index", filename, tmp.name, with_exceptions=False)

    def get_root_dir(self) -> str:
        return self.root_dir



import logging
import logging.config
import os
import pathlib
import tempfile

import yaml

__loggers = {}

source_dir = pathlib.Path(__file__).parent.resolve()


def load_logging_yaml(filepath: str = f"{source_dir}/logging.yaml"):
    with open(f'{filepath}', 'rt') as f:
        config = yaml.safe_load(f.read())
        try:
            if "file" in config["handlers"]:
                if not os.path.isdir('/tmp'):
                    config["handlers"]["file"]["filename"] = f"{tempfile.gettempdir()}/menschmachine.log"
            if "debug" in config["handlers"]:
                if not os.path.isdir('/tmp'):
                    config["handlers"]["debug"]["filename"] = f"{tempfile.gettempdir()}/menschmachine-debug.log"
        except:
            pass
        logging.config.dictConfig(config)


def get_logger(name: str = "menschmachine") -> logging.Logger:
    if name not in __loggers:
        logger = logging.getLogger(name)
        __loggers[name] = logger
    return __loggers[name]



import re
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Iterator, Optional, Callable

from bs4 import BeautifulSoup

from menschmachine import log, util


@dataclass
class LineRange:
    start: int
    end: int


class Operation(ABC):
    @abstractmethod
    def apply(self, text: str) -> str:
        raise NotImplementedError


@dataclass
class Remove(Operation):
    line_range: LineRange

    def apply(self, text: str) -> str:
        """
        Remove lines from a string based on the given line range.

        Args:
        text (str): The input string.
        start (int): The starting line number to remove (1-indexed).
        end (int): The ending line number to remove (1-indexed, inclusive).

        Returns:
        str: The modified string with specified lines removed.
        """
        # Split the text into lines
        lines = text.splitlines()

        # Adjust for 0-indexing
        start_index = self.line_range.start - 1
        end_index = self.line_range.end - 1

        # Validate input
        if start_index < 0:
            raise ValueError(f"RemoveOperation: Invalid line range: start line {self.line_range.start} is less than 1")
        if end_index >= len(lines):
            raise ValueError(
                f"RemoveOperation: Invalid line range: end line {self.line_range.end} exceeds file length of {len(lines)} lines")
        if start_index > end_index:
            raise ValueError(
                f"RemoveOperation: Invalid line range: start line {self.line_range.start} is greater than end line {self.line_range.end}")

        # Remove the specified lines
        del lines[start_index:end_index + 1]

        # Join the remaining lines and return
        return '\n'.join(lines)


def has_only_one_line(string):
    # Remove leading and trailing whitespace
    string = string.strip()

    # Check if there are any newline characters in the string
    return '\n' not in string


@dataclass
class Modify(Operation):
    line_number: int
    old_content: str
    new_content: str

    @staticmethod
    def _is_string_at_line(file_content: str, target_string: str, line_number: int) -> bool:
        """
        Check if a given string (potentially multiline) starts at the specified line number in the file content.

        Args:
        file_content (str): The content of the file as a string.
        target_string (str): The string to search for (can be multiline).
        line_number (int): The line number where the string should start (1-indexed).

        Returns:
        bool: True if the target string starts at the specified line number, False otherwise.
        """
        # Split the file content and target string into lines
        file_lines = file_content.splitlines()
        target_lines = target_string.splitlines()

        # Check if the starting line number is valid
        if line_number < 1 or line_number > len(file_lines):
            return False

        # Check if there are enough lines in the file to contain the target string
        if len(file_lines) - line_number + 1 < len(target_lines):
            return False

        # Compare the lines
        for i, target_line in enumerate(target_lines):
            if file_lines[line_number - 1 + i] != target_line:
                return False

        return True

    def apply(self, text: str) -> str:
        # check first if linenumber and old_content matches and can be found, if be strict by linenumber, otherwise do fuzzy search and replace
        if self.old_content is None or len(self.old_content.strip()) == 0:
            # means we are inserting at linenumber
            return self._singleline_replace(text)
        elif has_only_one_line(self.old_content) and self._is_string_at_line(text, self.old_content, self.line_number):
            return self._singleline_replace(text)
        else:
            return self._multiline_replace(text)

    def _singleline_replace(self, text: str) -> str:
        """
        Replace a specific line in a string representing a file with new content.

        Args:
        file_content (str): The original file content as a string.
        line_number (int): The line number to be replaced (1-indexed).
        new_content (str): The new content to replace the existing line.

        Returns:
        str: A new string with the specified line replaced by the new content.
        """
        # Split the file content into lines
        lines = text.splitlines()
        # adjust for zero indexing
        if self.line_number > 0:
            replace_index = self.line_number - 1
        else:
            replace_index = self.line_number

        # Validate input
        if replace_index < 0 or replace_index >= len(lines):
            raise ValueError(f"Invalid line number {self.line_number}/{replace_index}/{len(lines)}")

        # Replace the specified line with the new content
        lines[replace_index] = self.new_content

        # Join the lines and return
        return '\n'.join(lines)

    def _multiline_replace(self, text: str) -> str:
        """
        Replace old_content with new_content in a string representing a file,
        ignoring whitespace differences when searching for old_content.

        Args:
        file_content (str): The original file content as a string.
        old_content (str): The content to be replaced (whitespace insensitive).
        new_content (str): The new content to replace the old content.

        Returns:
        str: A new string with the old content replaced by the new content.
        """

        # first try the simple thing
        if self.old_content.strip() in text.strip():
            return text.strip().replace(self.old_content.strip(), self.new_content)

        # Remove leading whitespace from each line
        main_lines = [line.lstrip() for line in text.splitlines()]
        snippet_lines = [line.lstrip() for line in self.old_content.splitlines()]
        # Join the lines back into strings
        main_stripped = '\n'.join(main_lines)
        snippet_stripped = '\n'.join(snippet_lines)

        # Escape special regex characters in the snippet
        snippet_escaped = re.escape(snippet_stripped)

        # Replace newline characters with a regex pattern that allows for any whitespace
        snippet_pattern = snippet_escaped.replace('\\n', r'\s*\n\s*')

        # Search for the snippet in the main string
        match = re.search(snippet_pattern, main_stripped, re.MULTILINE)

        if match is None:
            log.get_logger().error(f"source: {text}, linenumber: {self.line_number}")
            log.get_logger().error(f"old_content: {self.old_content}")
            log.get_logger().error(f"new_content: {self.new_content}")
            raise ValueError("Content not found")
        else:
            # If found, replace the snippet with the replacement string
            result = re.sub(snippet_pattern, self.new_content, main_stripped, flags=re.MULTILINE)

            # Restore the original indentation
            result_lines = result.splitlines()
            indented_result = []
            indentation = 0
            for i, line in enumerate(result_lines):
                if i < len(main_lines):
                    # Use the indentation from the original main string
                    indentation = len(main_lines[i]) - len(main_lines[i].lstrip())
                    indented_result.append(' ' * indentation + line)
                else:
                    # For any additional lines, use the last known indentation
                    indented_result.append(' ' * indentation + line)

            return '\n'.join(indented_result)


@dataclass
class Delete(Modify):
    new_content: str = ""


@dataclass
class Add(Operation):
    after_line_number: int
    content: str

    def apply(self, text: str) -> str:
        """
          Insert new content at a specified line number in a string representing a file.

          Args:
          file_content (str): The original file content as a string.
          new_content (str): The new content to be inserted.
          line_number (int): The line number where the new content should be inserted (1-indexed).

          Returns:
          str: A new string with the new content inserted at the specified line number.
          """
        # Split the file content into lines
        lines = text.splitlines()

        insert_index = self.after_line_number

        # Validate input
        if insert_index < 0 or insert_index > len(lines):
            raise ValueError(f"Invalid line number: {insert_index}/{len(lines)}")

        # Insert the new content
        lines.insert(insert_index, self.content)
        return "\n".join(lines)


@dataclass
class Change:
    file_path: str
    intention: str
    operations: List[Operation] = field(default_factory=list)

    def apply(self, source: str) -> str:
        s = source
        for operation in self.operations:
            s = operation.apply(s)
        return s


@dataclass
class AppliedChanges:
    intention: str
    source: str
    filename: str


@dataclass
class Patch:
    changes: List[Change] = field(default_factory=list)

    def add_delete(self, line_number: int, content: str, file_path: str, intention: str):
        delete = Delete(line_number=line_number, old_content=content)
        change = Change(file_path=file_path, intention=intention)
        change.operations.append(delete)
        self.changes.append(change)

    def add_remove(self, start: int, end: int, file_path: str, intention: str):
        remove = Remove(line_range=LineRange(start, end))
        change = Change(file_path=file_path, intention=intention)
        change.operations.append(remove)
        self.changes.append(change)

    def add_modify(self, line_number: int, old_content: str, new_content: str, file_path: str, intention: str):
        modify = Modify(line_number=line_number, old_content=old_content, new_content=new_content)
        change = Change(file_path=file_path, intention=intention)
        change.operations.append(modify)
        self.changes.append(change)

    def add_add(self, after_line_number: int, content: str, file_path: str, intention: str):
        add = Add(after_line_number=after_line_number, content=content)
        change = Change(file_path=file_path, intention=intention)
        change.operations.append(add)
        self.changes.append(change)

    @staticmethod
    def from_dirty_xml(param: str):
        xml = Patch.xml_from_string(param)
        if xml.find("body"):
            patch = xml.find("body").find("patch")
        else:
            patch = xml
        return Patch.from_xml_object(patch)

    @staticmethod
    def xml_from_string(param: str) -> ET.XML:
        return util.xml_from_string(param)

    @staticmethod
    def from_xml(xml_string):
        try:
            root = ET.fromstring(xml_string)
            return Patch.from_xml_object(root)
        except:
            return Patch.from_dirty_xml(xml_string)

    @staticmethod
    def from_xml_object(root: ET.Element):
        patch = Patch()
        changes = root.find('changes')
        if changes is None and root.find("patch") is not None:
            return Patch.from_xml_object(root.find('patch'))
        if root.find("file_path") is not None:  # single change patch
            changes = [root]
        for change in changes:
            patch_change = Patch.parse_change(change)
            patch.changes.append(patch_change)
        return patch

    @staticmethod
    def parse_change(change):
        file_path = change.find('file_path').text
        intention = change.find('intention').text
        patch_change = Change(file_path=file_path, intention=intention)
        for child in change:
            operation = None
            if child.tag == "remove":
                operation = Patch.get_remove_op(child)
            elif child.tag == "modify":
                operation = Patch.get_modify_op(child)
            elif child.tag == "add":
                operation = Patch.get_add_op(child)
            if operation:
                patch_change.operations.append(operation)
            else:  # LLM generated some xml without explicitly specifying the operations, let's test what makes sense:
                if change.find("content") is not None:  # assuming add
                    operation = Patch.get_add_op(change)
                elif change.find("line_range") is not None:  # assuming delete
                    operation = Patch.get_remove_op(change)
                elif change.find("old_content") is not None and change.find(
                        "new_content") is not None:  # assuming modify
                    operation = Patch.get_modify_op(change)
                if operation is not None:
                    patch_change.operations.append(operation)
                    break  # there can be only one, because of no modify/remove/add tag
        return patch_change

    @staticmethod
    def get_remove_op(child):
        line_range = child.find('line_range')
        start = int(line_range.find('start').text)
        end = int(line_range.find('end').text)
        return Remove(line_range=LineRange(start, end))

    @staticmethod
    def get_add_op(child):
        after_line_number = int(child.find('after_line_number').text)
        content = child.find('content').text.strip()
        return Add(after_line_number=after_line_number, content=content)

    @staticmethod
    def get_modify_op(child):
        try:
            line_number = int(child.find('line_number').text)
        except ValueError:
            line_number = -1
        old_content = child.find('old_content').text
        new_content = child.find('new_content').text
        return Modify(line_number=line_number, old_content=old_content, new_content=new_content)

    def apply(self, get_file_content: Callable[[str], str | None]) -> Iterator[AppliedChanges]:
        for change in self.changes:
            path = change.file_path
            source = get_file_content(path)
            if source is None:
                log.get_logger().error(f"No sources for file {path} found")
                yield AppliedChanges(filename=path, source="", intention=change.intention)
            else:
                yield AppliedChanges(filename=path, source=change.apply(source), intention=change.intention)

    def apply_merged(self, get_file_content: Callable[[str], str | None]) -> Iterator[AppliedChanges]:
        changes_by_file = dict()
        for change in self.changes:
            path = change.file_path
            if path not in changes_by_file:
                changes_by_file[path] = list()
            changes_by_file[path].append(change)
        for k in changes_by_file.keys():
            changes = changes_by_file[k]
            source = get_file_content(k)
            if source is None:
                log.get_logger().debug(f"No sources for file {k} found")
                continue
            intentions = list()
            for change in changes:
                source = change.apply(source)
                intentions.append(change.intention)
            yield AppliedChanges(filename=k, source=source, intention=". ".join(intentions))


@dataclass
class Answer:
    description: str
    patch: Optional[Patch] = None

    def apply(self, get_file_content: Callable[[str], str | None]) -> Iterator[AppliedChanges]:
        if self.patch is not None:
            return self.patch.apply(get_file_content)

    def apply_merged(self, get_file_content: Callable[[str], str | None]) -> Iterator[AppliedChanges]:
        if self.patch is not None:
            return self.patch.apply_merged(get_file_content)

    @staticmethod
    def from_dirty_xml(param: str):
        xml_start = param.index("<")
        xml_end = param.rfind(">")
        xml_string = param[xml_start:xml_end + 1]
        soup = BeautifulSoup(xml_string, "xml")
        xml = ET.fromstring(str(soup))
        if xml.find("body"):
            answer = xml.find("body").find("answer")
        else:
            answer = xml
        return Answer._from_xml_object(answer)

    @staticmethod
    def from_xml(xml_string):
        try:
            root = ET.fromstring(xml_string)
            return Answer._from_xml_object(root)
        except:
            try:
                return Answer.from_dirty_xml(xml_string)
            except ValueError:
                return Answer(description=xml_string)

    @staticmethod
    def _from_xml_object(root):
        desc = None
        try:
            desc = root.find('description').text
        except:
            pass
        answer = Answer(description=desc)
        patch = root.find("patch")
        if patch is not None:
            answer.patch = Patch.from_xml_object(patch)
        return answer



import os
import time
from typing import Any, List

from menschmachine.llm.claude import ClaudeApi
from menschmachine.llm.lite import LiteLlmApi
from menschmachine.llm.oai import OAI
from menschmachine.llm.openrouter import OpenRouterApi
from menschmachine.llm.types import ApiInterface, FileRequestCallbackType, QuestionCallbackType, \
    SearchFilesCallbackType, \
    CreateFileCallbackType, ApiResponse, Agent, ApiCosts
from menschmachine.log import get_logger


class ApiProxy(ApiInterface):
    _instances = {}

    def __new__(cls):
        if cls not in cls._instances:
            cls._instances[cls] = super(ApiProxy, cls).__new__(cls)
            # noinspection PyProtectedMember
            cls._instances[cls]._initialize()
        return cls._instances[cls]

    def _initialize(self, wait_for_retry_secs=600):
        api_order = os.environ.get('MENSCHMACHINE_APIS', 'OpenRouterApi,ClaudeApi,OpenAI').split(',')
        api_mapping = {
            'OpenRouterApi': OpenRouterApi,
            'ClaudeApi': ClaudeApi,
            'LiteLlmApi': LiteLlmApi,
            'OpenAI': OAI,
        }

        self.apis: List = []
        for api_name in api_order:
            api_class = api_mapping.get(api_name.strip())
            if api_class:
                self.apis.append(api_class())
            else:
                raise ValueError(f"Warning: Unknown API '{api_name}' specified in MENSCHMACHINE_APIS")

        self.wait_for_retry_secs = wait_for_retry_secs

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        return self._get_api(model).estimate_costs(model, prompt_tokens, completion_tokens)

    def is_model_supported(self, model: str) -> bool:
        for api in self.apis:
            if api.is_model_supported(model):
                return True
        return False

    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 10) -> ApiResponse:
        return await self._get_api(model).agent_loop(uuid, prompt, agent_ids, model, max_tokens, temperature,
                                                     max_nr_of_loops)

    def ping(self) -> ApiResponse:
        return self._get_api().ping()

    def _get_supported_apis(self, model: str) -> list[ApiInterface]:
        apis = list()
        for api in self.apis:
            if model is None or api.is_model_supported(model):
                apis.append(api)
        return apis

    def _get_api(self, model: str = None) -> ApiInterface:
        if "UNIVERSAL_MODEL" in os.environ:
            model = os.environ["UNIVERSAL_MODEL"]
        if model is None:
            if "DEFAULT_MODEL" in os.environ:
                model = os.environ["UNIVERSAL_MODEL"]
            else:
                model = "haiku"
        apis = self._get_supported_apis(model)
        for api in apis:
            if self.wait_for_retry_secs + api.failed_at > time.time():
                continue  # means api is marked as 'failing'
            elif self.wait_for_retry_secs + api.success_at > time.time():
                # recently pinged successfully, no ping() needded
                return api
            else:
                try:
                    api.ping()
                    api.success_at = time.time()
                    api.failed_at = -1
                    api.failed_reason = None
                    return api
                except Exception as e:
                    get_logger().info(f"Api {api.__class__.__name__} is failing. Reason: {e}")
                    get_logger().debug(f"Api {api.__class__.__name__} is failing", exc_info=True)
                    api.failed_at = time.time()
                    api.failed_reason = e
                    api.success_at = -1
        raise ValueError(f"No api found supporting {model} which is not marked as failing")

    def ask(self,
            prompt: str,
            model: str = None,
            max_tokens: int = 4096,
            temperature=0.0,
            assistant_msg=None) -> ApiResponse:
        response = self._get_api(model).ask(prompt, model, max_tokens, temperature, assistant_msg)
        get_logger().info(f"{response.costs}")
        return response

    async def with_tools_loop(self,
                              prompt: str,
                              file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType,
                              model: str,
                              max_tokens: int = 4096,
                              temperature: float = 0.1) -> ApiResponse:
        response = await self._get_api(model).with_tools_loop(prompt, file_content_callback,
                                                              follow_up_question_callback,
                                                              search_files_callback, create_file_callback, model,
                                                              max_tokens, temperature)
        get_logger().info(f"{response.costs}")
        return response

    async def loop(self, uuid: str, model: str, agent_map: dict[str, Agent], nr_of_loops: int, max_nr_of_loops: int,
                   messages: list, temperature: float, costs: ApiCosts) -> Any:
        return await self._get_api(model).loop(uuid, model, agent_map,
                                               nr_of_loops, max_nr_of_loops,
                                               messages, temperature, costs)


# Create a single instance of ApiProxy
api_proxy = ApiProxy()



import json
from json import JSONDecodeError

from menschmachine.llm.proxy import ApiProxy
from menschmachine.llm.types import ApiResponse
from menschmachine.log import get_logger


def get_fix_json_prompt(json_str):
    return f'''The following json cannot be parsed using json.loads() on python. 
    Return an answer that is directly parsable with json.loads(). Pay especially attention to string quoting problems.
    If you don't know how to return a parseable json string, return an empty array.
    IMPORTANT: only respond with the json, no explanations or anything else. If the json is already valid, just echo it.
```json
{json_str}
```
'''


def _normalize_json(json_str):
    json_str = json_str.strip()
    first_index_of_square_bracket = json_str.find('[')
    first_index_of_curly_bracket = json_str.find('{')

    if first_index_of_square_bracket > -1 and (
            first_index_of_curly_bracket == -1 or first_index_of_square_bracket < first_index_of_curly_bracket):
        json_start = first_index_of_square_bracket
        json_end = json_str.rfind(']')
    elif first_index_of_curly_bracket > -1:
        json_start = first_index_of_curly_bracket
        json_end = json_str.rfind('}')
    else:
        json_start = -1
        json_end = -1

    if json_start > -1 and json_end > -1:
        return json_str[json_start:json_end + 1]
    else:
        return json_str


def to_json(json_str: str, fallback_value=None) -> object:
    try:
        return json.loads(_normalize_json(json_str))
    except JSONDecodeError:
        return _fix_json(json_str, fallback_value)


def _fix_json(json_str: str, fallback_value=None) -> object:
    get_logger().debug(f"Fixing {json_str}")
    prompt = get_fix_json_prompt(json_str)
    fixed_json = ApiProxy().ask(prompt, model="haiku").message
    try:
        return json.loads(_normalize_json(fixed_json))
    except JSONDecodeError:
        get_logger().exception(
            f"Fixed json - could not be parsed: {_normalize_json(fixed_json)}")
        return fallback_value


class MM(object):
    @staticmethod
    def json(json_str: str):
        return to_json(json_str)

    @staticmethod
    def ask(query: str, model: str = None) -> ApiResponse:
        return ApiProxy().ask(query, model=model)

    @staticmethod
    def ask_for_json(query: str, model: str = None) -> object:
        return to_json(ApiProxy().ask(query, model=model).message)


if __name__ == "__main__":
    print(MM.json("Here is the json: {'x': 1, 'y': 2}"))



import json
import os
from typing import List, Any

import litellm
from openai import OpenAI, RateLimitError
from openai.types.chat import ChatCompletion, ChatCompletionMessageToolCall

from menschmachine.event import events
from menschmachine.llm.types import FileRequestCallbackType, QuestionCallbackType, CreateFileCallbackType, \
    SearchFilesCallbackType, \
    ApiInterface, ApiResponse, MaxNumberOfLoops, ApiCosts, Agent, ModelRegistry
from menschmachine.log import get_logger
from menschmachine.util import symetric_dict

MODEL_REGISTRY = ModelRegistry("haiku")

# Register models with their aliases
MODEL_REGISTRY.register("litellm-haiku", ["haiku", "claude-haiku"], 0.25, 1.25, "anthropic/claude-3-haiku-20240307")
MODEL_REGISTRY.register("litellm-sonnet", ["sonnet", "claude-sonnet"], 3, 15, "anthropic/claude-3-5-sonnet-20240620")
MODEL_REGISTRY.register("litellm-opus", ["opus", "claude-opus"], 15, 75, "anthropic/claude-3-opus-20240229")
MODEL_REGISTRY.register("litellm-gpt-4o", ["gpt4o", "gpt4", "gpt-4o"], 2.5, 10, "openai/gpt-4o")
MODEL_REGISTRY.register("litellm-o1-mini", ["gpt4o-mini", "gpt4-mini"], 3, 12)
MODEL_REGISTRY.register("litellm-gemini-pro", ["gemini-pro"], 1.25, 5)
MODEL_REGISTRY.register("litellm-gemini-flash", ["gemini-flash"], 0.075, 0.3, "gemini/gemini-1.5-flash-latest")
MODEL_REGISTRY.register("litellm-o1-preview-2024-09-12", ["o1", "gpt-o1", "o1-preview"], 15, 60)


def raw(messages: list[dict],
        tools: list[dict] = None,
        tool_choice: dict | str | None = None,
        model: str = None,
        max_tokens: int = 4000,
        temperature=0.0,
        ) -> ChatCompletion:
    client = create_client()
    model = MODEL_REGISTRY.get_model(model)
    official_name = MODEL_REGISTRY.get_official_name(model)
    assert litellm.supports_function_calling(model=official_name)

    if temperature is None:
        temperature = 0.0
    try:
        result = None
        if tools is None:
            result = client.chat.completions.create(
                max_tokens=max_tokens,
                temperature=temperature,
                messages=messages,
                model=model,
            )
        else:
            if tool_choice is None:
                result = client.chat.completions.create(
                    max_tokens=max_tokens,
                    temperature=temperature,
                    messages=messages,
                    tools=tools,
                    model=model,
                )
            else:
                result = client.chat.completions.create(
                    max_tokens=max_tokens,
                    temperature=temperature,
                    messages=messages,
                    tools=tools,
                    tool_choice=tool_choice,
                    model=model,
                )
        return result

    except RateLimitError as e:
        headers = e.response.headers
        rate_limit_info = {
            'x-ratelimit-limit-requests': headers.get('x-ratelimit-limit-requests'),
            'x-ratelimit-remaining-requests': headers.get('x-ratelimit-remaining-requests'),
            'x-ratelimit-reset-requests': headers.get('x-ratelimit-reset-requests'),
            'x-ratelimit-limit-tokens': headers.get('x-ratelimit-limit-tokens'),
            'x-ratelimit-remaining-tokens': headers.get('x-ratelimit-remaining-tokens'),
            'x-ratelimit-reset-tokens': headers.get('x-ratelimit-reset-tokens'),
            'retry-after': headers.get('retry-after'),
        }

        # Create a log message to show exceeded limits
        exceeded_limits = []

        # Check for exceeded limits
        if int(rate_limit_info['x-ratelimit-remaining-tokens']) == 0:
            exceeded_limits.append('Tokens limit exceeded')

        if int(rate_limit_info['x-ratelimit-remaining-requests']) == 0:
            exceeded_limits.append('Requests limit exceeded')

        # Prepare the log message
        if exceeded_limits:
            get_logger().error(f"Rate limit exceed: [{', '.join(exceeded_limits)}]")
            exceeded_info = {key: value for key, value in rate_limit_info.items() if value is not None}
            get_logger().debug(f"Rate limits exceeded: {exceeded_info}")
        else:
            get_logger().info("No rate limits exceeded.")
        raise e


def create_client():
    client = OpenAI(
        api_key=os.environ["LITELLM_API_KEY"],
        base_url="http://code.thefamouscat.com:4000/api/v1"
    )
    return client


class LiteLlmApi(ApiInterface):

    def __init__(self):
        super().__init__()

    def is_model_supported(self, model: str) -> bool:
        return MODEL_REGISTRY.is_supported(model)

    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 10) -> ApiResponse:

        costs = self.get_api_costs(model, 0, 0)
        agent_map = self.get_agent_map(agent_ids)
        tools = self.tools_from_agent_map(agent_map)
        messages = [
            {
                "role": "user",
                "content": prompt,
            }
        ]
        events.emit(f"/mm/agent/init/{uuid}", uuid, "openrouter", symetric_dict(prompt, model, tools, messages))
        return await self.loop(uuid, model, agent_map, 0, max_nr_of_loops, messages, temperature, costs)

    def tools_from_agent_map(self, agent_map):
        tools = []
        for agent_id in agent_map:
            agent = agent_map[agent_id]
            properties, required_properties = self.properties_for_agent(agent)
            input_schema = self.input_schema_from_properties(properties, required_properties)
            tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": agent.name,
                        "description": agent.description,
                        "parameters": input_schema
                    }
                }
            )
        return tools

    def ping(self) -> ApiResponse:
        return self.ask("ping", MODEL_REGISTRY.get_long_name("haiku"))

    def ask(self,
            prompt: str,
            model: str = None,
            max_tokens: int = 4096,
            temperature=0.0,
            assistant_msg=None) -> ApiResponse:

        messages = [{"role": "user", "content": prompt}]
        if assistant_msg:
            messages.append({"role": "assistant", "content": assistant_msg})

        model = MODEL_REGISTRY.get_model(model)
        response = raw(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature
        )
        costs = self.get_api_costs(model, response.usage.prompt_tokens, response.usage.completion_tokens)
        answer = response.choices[0].message.content
        return ApiResponse(message=answer, error=None, costs=costs)

    async def with_tools_loop(self, prompt: str,
                              file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType,
                              model: str,
                              max_tokens: int = 4000,
                              temperature: float = 0.1) -> ApiResponse:
        raise NotImplementedError("deprecated, use loop()")

    async def loop(self,
                   uuid: str,
                   model: str,
                   agent_map: dict[str, Agent],
                   nr_of_loops: int,
                   max_nr_of_loops: int,
                   messages: list,
                   temperature: float,
                   costs: ApiCosts) -> Any:

        async def append_next_step(tool_result: Any, _response: str, function_name: str, tool_call_id: str):
            messages.append({"role": "assistant", "content": _response})
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call_id,
                    "name": function_name,
                    "content": json.dumps(tool_result),
                })
            events.emit(f"/mm/agent/step/{uuid}", uuid, "openrouter", messages)

        nr_of_loops += 1
        get_logger().debug(f"Loop number: {nr_of_loops}")
        if nr_of_loops > max_nr_of_loops:
            get_logger().warning("Max number of loops reached")
            raise MaxNumberOfLoops()
        tools = self.tools_from_agent_map(agent_map)
        try:
            response = raw(
                model=model,
                messages=messages,
                tools=tools,
                temperature=temperature,
                max_tokens=4096
            )
            costs.add(self.get_api_costs(model, response.usage.prompt_tokens, response.usage.completion_tokens))
            choice = response.choices[0]
            message = choice.message
            get_logger().debug(f"Choice: {choice}")
            if choice.finish_reason == 'tool_calls' or (message.tool_calls is not None and len(message.tool_calls) > 0):

                tool_calls: List[ChatCompletionMessageToolCall] = message.tool_calls
                tool_call_results = list[dict]()
                for tool_call in tool_calls:
                    function_args = tool_call.function.arguments
                    tool_use_id = tool_call.id
                    tool_name = tool_call.function.name
                    get_logger().info(f"Tool requested: {tool_name}")
                    get_logger().debug(f"Tool request messsage: {message}")
                    if tool_name not in agent_map:
                        raise ValueError(f"Unknown tool requested: {tool_name}")
                    # noinspection DuplicatedCode
                    try:
                        await agent_map[tool_name].call(uuid,
                                                        json.loads(function_args),
                                                        response,
                                                        tool_use_id,
                                                        append_next_step)
                    except Exception as e:
                        self.handle_tool_call_exception(message, tool_name, e)
                    await self.loop(
                        uuid,
                        model,
                        agent_map,
                        nr_of_loops,
                        max_nr_of_loops,
                        messages,
                        temperature,
                        costs)
                messages.extend(tool_call_results)
                return await self.loop(
                    uuid,
                    model,
                    agent_map,
                    nr_of_loops,
                    max_nr_of_loops,
                    messages,
                    temperature,
                    costs)
            elif choice.finish_reason == 'end_turn' or choice.finish_reason.lower() == 'stop':
                get_logger().info("Agent loop completed successfully")
                events.emit(f"/mm/agent/completed/{uuid}", uuid, "openrouter", symetric_dict(response, costs))
                return ApiResponse(message=message.content, error=None, costs=costs)
            else:
                get_logger().warning(f"Unexpected stop reason: {choice.finish_reason}")
                return ApiResponse(message=message.content, error=None, costs=costs)
        except Exception as e:
            events.emit(f"/mm/agent/error/{uuid}", uuid, "openrouter", symetric_dict(e, costs))
            get_logger().error(f"Error in agent loop: {str(e)}")
            raise

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        def costs(nr_tokens, price_per_million: float):
            return (nr_tokens * (price_per_million / (1000. * 1000.))) / 100.  # in cent

        in_price, out_price = MODEL_REGISTRY.get_prices(model)
        return costs(prompt_tokens, in_price) + costs(completion_tokens, out_price)



import json
import os
import tempfile
from abc import ABC, abstractmethod, ABCMeta
from dataclasses import dataclass, field
from inspect import isawaitable
from typing import Callable, Awaitable, Optional, Any, Coroutine, List, Dict, Set, Tuple

from dataclasses_json import dataclass_json
from fastapi.encoders import jsonable_encoder

from menschmachine.event import events
from menschmachine.log import get_logger
from menschmachine.util import symetric_dict

FileRequestCallbackType = Callable[[list[str]], Awaitable[dict[str, str]]]
QuestionCallbackType = Callable[[str], Awaitable[str]]
CreateFileCallbackType = Callable[[str, str], Awaitable[str]]
SearchFilesCallbackType = Callable[[str], Awaitable[list[str]]]


@dataclass
class FileQueryResponse:
    answer: str
    files: dict[str, str] = field(default_factory=dict)


@dataclass
class ApiError:
    message: str


@dataclass
class RateLimitError(ApiError):
    pass


@dataclass
class ApiCosts:
    tokens_in: int
    tokens_out: int
    model: str
    estimated_cost_in_us_cent: float = field(repr=False)
    price: Optional[float] = 0.

    def add(self, costs):
        self.tokens_in += costs.tokens_in
        self.tokens_out += costs.tokens_out
        self.estimated_cost_in_us_cent += costs.estimated_cost_in_us_cent
        self.price = round(self.estimated_cost_in_us_cent, 4)

    @classmethod
    def from_json(cls, api_costs_json: dict) -> 'ApiCosts':
        """

        Args:
            api_costs_json:

        {'estimated_cost_in_us_cent': 5.8822499999999995e-05, 'model': 'claude-3-haiku-20240307', 'price': 0.0001,
         'tokens_in': 15469, 'tokens_out': 1612}
        Returns:

        """

        return ApiCosts(estimated_cost_in_us_cent=api_costs_json['estimated_cost_in_us_cent'],
                        model=api_costs_json['model'],
                        tokens_out=api_costs_json['tokens_out'],
                        tokens_in=api_costs_json['tokens_in'],
                        price=api_costs_json['price'])


@dataclass_json
@dataclass
class ApiResponse:
    message: str
    error: ApiError | None
    costs: ApiCosts
    files: Optional[dict[str, str]] = None


@dataclass
class ToolProperty:
    name: str
    description: str
    required: bool = True
    type: str = "string"


CollectResultCallable = Callable[[Any, Any, str, str], Coroutine[Any, Any, None]]


class Agent:

    def __init__(self,
                 name: str,
                 description: str,
                 properties: list[ToolProperty],
                 callback: Callable[[str, dict], Any] = None):
        super().__init__()
        self.name = name
        self.description = description
        self.properties = properties
        self.callback = callback

    async def call(self,
                   uuid: str,
                   msg: dict,
                   response: Any,
                   tool_use_id: str | None,
                   collect_result: CollectResultCallable | None = None) -> Any:
        get_logger().debug(f"Agent {self.name} called with {msg}")
        result = self.callback(uuid, msg)
        if isawaitable(result):
            result = await result
        events.emit(f"/mm/agent/call/result/{uuid}", uuid, None,
                    symetric_dict(result, self.name, msg, response, tool_use_id))
        get_logger().debug(f"Agent {self.name} result: {result}")
        if collect_result is not None:
            return await collect_result(result, response, self.name, tool_use_id)
        else:
            return result


class AgentRepository:
    _instance = None

    def __init__(self):
        super().__init__()
        self.agents: dict[str, Agent] = {}

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(AgentRepository, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def register(self, agent_id: str, agent: Agent):
        self.agents[agent_id] = agent

    def get(self, agent_id: str) -> Agent:
        return self.agents[agent_id]


agent_repository: AgentRepository = AgentRepository()


class MaxNumberOfLoops(Exception):
    pass


class ApiInterface(ABC, metaclass=ABCMeta):

    def __init__(self):
        self.failed_at = -1
        self.success_at = -1
        self.failed_reason: Exception | None = None

    @abstractmethod
    def ping(self) -> ApiResponse:
        raise NotImplementedError()

    @abstractmethod
    def is_model_supported(self, model: str) -> bool:
        return False

    @abstractmethod
    def ask(self,
            prompt: str,
            model: str = None,
            max_tokens: int = 4096,
            temperature=0.0,
            assistant_msg=None
            ) -> ApiResponse:
        raise NotImplementedError()

    @abstractmethod
    async def with_tools_loop(
            self,
            prompt: str,
            file_content_callback: FileRequestCallbackType,
            follow_up_question_callback: QuestionCallbackType,
            search_files_callback: SearchFilesCallbackType,
            create_file_callback: CreateFileCallbackType,
            model: str,
            max_tokens: int = 4096,
            temperature: float = 0.1,
    ) -> ApiResponse:
        raise NotImplementedError()

    @abstractmethod
    async def agent_loop(self,
                         uuid: str,
                         prompt: str,
                         agent_ids: list[str],
                         model: str,
                         max_tokens: int = 4096,
                         temperature: float = 0.1,
                         max_nr_of_loops: int = 10
                         ) -> ApiResponse:
        raise NotImplementedError()

    def get_api_costs(self, model, tokens_in, tokens_out) -> ApiCosts:
        return ApiCosts(tokens_in=tokens_in,
                        tokens_out=tokens_out,
                        model=model,
                        estimated_cost_in_us_cent=self.estimate_costs(model,
                                                                      tokens_in,
                                                                      tokens_out))

    @abstractmethod
    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        raise NotImplementedError()

    # noinspection PyMethodMayBeStatic
    def input_schema_from_properties(self, properties, required_properties):
        input_schema = {
            "type": "object",
            "properties": properties,
            "required": required_properties
        }
        return input_schema

    # noinspection PyMethodMayBeStatic
    def properties_for_agent(self, agent):
        required_properties = []
        properties = {}
        for prop in agent.properties:
            properties[prop.name] = {
                "type": prop.type,
                "description": prop.description
            }
            if prop.required:
                required_properties.append(prop.name)
        return properties, required_properties

    # noinspection PyMethodMayBeStatic
    def get_agent_map(self, agent_ids: list[str]) -> dict[str, Agent]:
        return {agent_repository.get(agent_id).name: agent_repository.get(agent_id) for agent_id in agent_ids}

    # noinspection PyMethodMayBeStatic
    def handle_tool_call_exception(self, message, tool_name, e: Exception):
        get_logger().error(f"Error calling agent {tool_name}", exc_info=True)
        dirpath = tempfile.mkdtemp()
        get_logger().info(f"Saving debug info to {dirpath}")
        with open(os.path.join(dirpath, f"{tool_name}-msg-debug.json"), "w") as f:
            try:
                f.write(message.to_json())
            except:
                f.write(json.dumps(jsonable_encoder(message)))
            f.flush()
        raise e

    @abstractmethod
    async def loop(self,
                   uuid: str,
                   model: str,
                   agent_map: dict[str, Agent],
                   nr_of_loops: int,
                   max_nr_of_loops: int,
                   messages: list,
                   temperature: float,
                   costs: ApiCosts) -> Any:
        raise NotImplementedError()


class ModelInfo:
    def __init__(self, long_name: str, short_names: List[str], input_price: float, output_price: float,
                 official_name: str):
        self.long_name = long_name
        self.short_names = short_names
        self.input_price = input_price
        self.output_price = output_price
        self.official_name = official_name


class ModelRegistry:
    def __init__(self, default_model: str):
        self.models: Dict[str, ModelInfo] = {}
        self.short_to_long: Dict[str, str] = {}
        self.default_model = default_model

    def register(self, long_name: str, short_names: List[str], input_price: float, output_price: float,
                 official_name: str = None):
        model_info = ModelInfo(long_name, short_names, input_price, output_price, official_name)
        self.models[long_name] = model_info
        for short_name in short_names:
            self.short_to_long[short_name.lower()] = long_name

    def get_long_name(self, name: str) -> str:
        return self.short_to_long.get(name.lower(), name)

    def is_supported(self, name: str) -> bool:
        long_name = self.get_long_name(name)
        return long_name in self.models

    def get_all_names(self) -> Set[str]:
        return set(self.models.keys()) | set(self.short_to_long.keys())

    def get_prices(self, name: str) -> Tuple[float, float]:
        long_name = self.get_long_name(name)
        if long_name in self.models:
            model_info = self.models[long_name]
            return model_info.input_price, model_info.output_price
        return -1.0, -1.0  # unknown, not found

    def get_model(self, model: str) -> str:
        if "UNIVERSAL_MODEL" in os.environ:
            model = os.environ["UNIVERSAL_MODEL"]
        if model is None:
            model = self.get_long_name(self.default_model)
        return self.get_long_name(model)

    def get_supported_models(self) -> Set[str]:
        return self.get_all_names()

    def get_official_name(self, model):
        model = self.get_model(model)
        for m in self.models:
            if m == model and self.models[m].official_name is not None:
                return self.models[m].official_name
        return model



import json
from typing import List, Dict, Any
from xml.etree import ElementTree

from menschmachine.patch import Patch


class Task:
    def __init__(self, name: str,
                 description: str = None,
                 estimation_in_minutes: int = None,
                 estimation_in_hours: int = None):
        self.name = name
        self.description = description
        self.estimation_in_hours = estimation_in_hours
        self.estimation_in_minutes = estimation_in_minutes

    def __repr__(self):
        return f"Task(name='{self.name}')"


class Step:
    def __init__(self, name: str, tasks: List[Task]):
        self.name = name
        self.tasks = tasks

    def __repr__(self):
        return f"Step(name='{self.name}', tasks={len(self.tasks)})"

    def add_task(self, task):
        self.tasks.append(task)


class Project:
    def __init__(self, name: str, steps: List[Step]):
        self.name = name
        self.steps = steps

    def __repr__(self):
        return f"Project(name='{self.name}', steps={len(self.steps)})"

    def add_step(self, step):
        self.steps.append(step)

    def to_xml(self) -> str:
        root = ElementTree.Element("project")
        name = ElementTree.SubElement(root, "name")
        name.text = self.name
        steps_elem = ElementTree.SubElement(root, "steps")
        for s in self.steps:
            step_elem = ElementTree.SubElement(steps_elem, "step")
            name = ElementTree.SubElement(step_elem, "name")
            name.text = s.name
            tasks_elem = ElementTree.SubElement(step_elem, "tasks")
            for t in s.tasks:
                task_elem = ElementTree.SubElement(tasks_elem, "task")
                task_name = ElementTree.SubElement(task_elem, "name")
                task_name.text = t.name
                if t.description:
                    task_desc = ElementTree.SubElement(task_elem, "description")
                    task_desc.text = t.description
                if t.estimation_in_minutes:
                    task_m = ElementTree.SubElement(task_elem, "estimation_in_minutes")
                    task_m.text = str(t.estimation_in_minutes)
                if t.estimation_in_hours:
                    task_h = ElementTree.SubElement(task_elem, "estimation_in_hours")
                    task_h.text = str(t.estimation_in_hours)
        return ElementTree.tostring(root, encoding="unicode")

    @classmethod
    def merge(cls, name: str, projects: list['Project']) -> 'Project':
        merged = Project(name=name, steps=[])
        for p in projects:
            for step in p.steps:
                merged.add_step(step)
        return merged


def parse_json(json_data: Dict[str, Any], name_elem: str = "project") -> Project:
    project_name = json_data[name_elem]
    steps = []

    for step_data in json_data['steps']:
        tasks = [Task(task['name'], task['description']) for task in step_data['tasks']]
        step = Step(step_data['name'], tasks)
        steps.append(step)

    return Project(project_name, steps)


# Function to load JSON from file
def load_json_from_file(file_path: str) -> Dict[str, Any]:
    with open(file_path, 'r') as file:
        return json.load(file)


def json_from_file(file_path: str) -> Project:
    json_data = load_json_from_file(file_path)
    return parse_json(json_data)


def parse_xml(xml_string: str) -> Project:
    root = Patch.xml_from_string(xml_string)

    project_name = root.find('name').text
    project = Project(project_name, list())

    for step_elem in root.find('steps').findall('step'):
        step_name = step_elem.find('name').text
        step = Step(step_name, list())

        for task_elem in step_elem.find('tasks').findall('task'):
            task_name = task_elem.find('name').text
            task_description = None
            hours = None
            minutes = None
            description_elem = task_elem.find('description')
            if description_elem is not None:
                task_description = description_elem.text
            hours_elem = task_elem.find('estimation_in_hours')
            if hours_elem is not None:
                hours = int(hours_elem.text)
            minutes_elem = task_elem.find('estimation_in_minutes')
            if minutes_elem is not None:
                minutes = int(minutes_elem.text)
            task = Task(task_name, description=task_description, estimation_in_hours=hours,
                        estimation_in_minutes=minutes)
            step.add_task(task)

        project.add_step(step)

    return project


def xml_from_file(file_path: str) -> Project:
    with open(file_path, 'r') as file:
        return parse_xml(file.read())


class StepCallback:
    def after_apply(self, affected_files: list[str]) -> str:
        pass

    def after_review(self, affected_files: list[str], task_name: str, task_nr: int, total_nr_of_tasks: int) -> None:
        pass



import inspect
import re
import xml.etree.ElementTree as ET

from bs4 import BeautifulSoup


def symetric_dict(*args):
    caller_locals = inspect.currentframe().f_back.f_locals
    return {name: value for name, value in caller_locals.items() if value in args}


def xml_from_string(param: str) -> ET.XML:
    xml_start = param.index("<")
    xml_end = param.rfind(">")
    xml_string = param[xml_start:xml_end + 1]
    xml_string = fix_cdata_sections(xml_string)
    soup = BeautifulSoup(xml_string, "xml")
    xml = ET.fromstring(str(soup))
    return xml


def fix_cdata_sections(xml_string: str) -> str:
    def replace_cdata(match):
        content = match.group(1)
        if content.endswith(']]') or content.endswith(']>'):
            content = content[:-2]
        elif content.endswith(']'):
            content = content[:-1]
        return f'<![CDATA[{content}]]>'

    pattern = r'<!\[CDATA\[(.*?)(?:\]\]*>)'
    fixed_xml = re.sub(pattern, replace_cdata, xml_string, flags=re.DOTALL)
    return fixed_xml



import os
from enum import Enum

code_suffixes = ["ts", "js", "java", "py", "c", "cc", "rb", "kt", "html"]
with open(os.path.dirname(os.path.realpath(__file__)) + '/code-file-extensions.txt') as f:
    code_suffixes.extend([x.strip() for x in f.read().splitlines()])

config_suffixes = ["properties", "yaml", "cfg", "conf", "xml"]
documentation_suffixes = ["md", "txt", "rst", "adoc"]
script_suffixes = ["sh", "bat", "cmd"]
asset_suffixes = ['jpg', 'jpeg', 'jpe', 'jif', 'jfif', 'jfi',  # JPEG images
                  'png',  # PNG images
                  'gif',  # GIF images
                  'webp',  # WebP images
                  'tiff', 'tif',  # TIFF images
                  'bmp', 'dib',  # BMP images
                  'jp2', 'j2k', 'jpf', 'jpx', 'jpm', 'mj2',  # JPEG 2000 images
                  'svg', 'svgz',  # SVG images
                  'ico', 'cur',  # Icon files
                  'heic', 'heif',  # High Efficiency Image Format
                  'psd',  # Adobe Photoshop files
                  'ai',  # Adobe Illustrator files
                  'raw', 'arw', 'cr2', 'nrw', 'k25',  # RAW images
                  'webp',  # WebP images
                  'tga',  # Truevision TGA
                  'pcx',  # PCX images
                  'ttf',  # TrueType Font
                  'otf',  # OpenType Font
                  'woff',  # Web Open Font Format
                  'woff2',  # Web Open Font Format 2
                  'eot',  # Embedded OpenType
                  'pfb',  # Type 1 Font (Printer Font Binary)
                  'pfm',  # Printer Font Metrics
                  'afm',  # Adobe Font Metrics
                  'tfm',  # TeX Font Metrics
                  'fon',  # Windows Font File
                  'fnt',  # Windows Font File
                  'bdf',  # Bitmap Distribution Format
                  'pcf',  # Portable Compiled Format
                  'snf',  # Server Normal Format
                  'otb',  # OpenType Bitmap
                  'cff',  # Compact Font Format
                  'psf',  # PC Screen Font
                  'ttc',  # TrueType Collection
                  'sfd',  # FontForge Spline Font Database
                  'ufo',  # Unified Font Object
                  'pfa',  # Printer Font ASCII
                  'dfont'  # Mac OS X Data Fork Font
                  ]
binary_suffixes = ["exe", "dll", "so", "lib", "bin"]
test_suffixes = ["test", "spec"]
data_suffixes = ["csv", "tsv", "json", "yml", "xls", "xlsx", "ods", "tsv"]


class FileType(Enum):
    CODE = "CODE"
    CONFIG = "CONFIG"
    DOCUMENTATION = "DOCUMENTATION"
    SPEC = "SPEC"
    SCRIPT = "SCRIPT"
    ASSET = "ASSET"
    BINARY = "BINARY"
    TEST = "TEST"
    HELPER = "HELPER"
    DATA = "DATA"
    UNKNOWN = "UNKOWN"


def get_file_type(filename: str) -> FileType:
    _, suffix = os.path.splitext(filename)
    if suffix.startswith("."):
        suffix = suffix[1:]
    suffix = suffix.lower()
    if suffix in documentation_suffixes:
        return FileType.DOCUMENTATION
    if suffix in script_suffixes:
        return FileType.SCRIPT
    if suffix in asset_suffixes:
        return FileType.ASSET
    if suffix in binary_suffixes:
        return FileType.BINARY
    if suffix in test_suffixes:
        return FileType.TEST
    if suffix in data_suffixes:
        return FileType.DATA
    if suffix in config_suffixes:
        return FileType.CONFIG
    if suffix in code_suffixes:
        return FileType.CODE
    return FileType.UNKNOWN



from typing import Any

from langchain_anthropic import ChatAnthropic
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_google_genai import ChatGoogleGenerativeAI

from menschmachine.llm.types import ApiInterface, Agent, ApiCosts, ApiResponse, FileRequestCallbackType, \
    QuestionCallbackType, SearchFilesCallbackType, CreateFileCallbackType, ModelRegistry

MODEL_REGISTRY = ModelRegistry("haiku")
MODEL_REGISTRY.register("claude-3-haiku-20240307", ["haiku", "claude-haiku"], 0.25, 1.25)
MODEL_REGISTRY.register("claude-3-5-sonnet-20240620", ["sonnet", "claude-sonnet"], 3, 15)
MODEL_REGISTRY.register("gemini-1.5-flash", ["gemini-flash"], 0.075, 0.3)


def raw(messages: list[BaseMessage] | list[dict[str, str]],
        tools: list[dict] = None,
        tool_choice: dict = None,
        model: str = "haiku",
        temperature=0.0,
        ) -> Any:
    client = L4ngCh4in()
    chat = client.create_chat(MODEL_REGISTRY.get_model(model))
    if temperature is None:
        temperature = 0.0
    if tools is None or len(tools) == 0:
        return raw_without_tools(chat, messages, temperature)
    else:
        if tool_choice is None:
            return raw_with_tools(chat, messages, temperature, tools)
        else:
            return raw_with_tools_and_choices(chat, messages, temperature, tool_choice, tools)


def raw_with_tools_and_choices(chat, messages: list[BaseMessage], temperature, tool_choice, tools) -> BaseMessage:
    return chat.invoke(
        input=messages,
        tools=tools,
    )


def raw_with_tools(chat: BaseChatModel, messages: list[BaseMessage], temperature, tools) -> BaseMessage:
    return chat.invoke(
        input=messages,
        tools=tools,
    )


def raw_without_tools(chat: BaseChatModel, messages: list[BaseMessage], temperature: float) -> BaseMessage:
    return chat.invoke(input=messages)


class L4ngCh4in(ApiInterface):

    # noinspection PyMethodMayBeStatic
    def create_chat(self, model: str) -> BaseChatModel:
        if "claude" in model:
            return ChatAnthropic(model_name=model)
        elif "gemini" in model:
            return ChatGoogleGenerativeAI(model=model)
        else:
            raise ValueError(model)

    def ping(self) -> ApiResponse:
        return self.ask("Ping!")

    def _get_api_costs_from_response(self, model, response):
        in_tokens = response.usage_metadata["input_tokens"]
        out_tokens = response.usage_metadata["output_tokens"]
        costs = ApiCosts(tokens_in=in_tokens,
                         tokens_out=out_tokens,
                         model=model,
                         estimated_cost_in_us_cent=self.estimate_costs(model,
                                                                       in_tokens,
                                                                       out_tokens))
        return costs

    def is_model_supported(self, model: str) -> bool:
        raise NotImplementedError()

    def ask(self, prompt: str, model: str = None, max_tokens: int = 4096, temperature=0.0,
            assistant_msg=None) -> ApiResponse:
        model = MODEL_REGISTRY.get_official_name(model)
        chat = self.create_chat(model)
        response = chat.invoke(prompt)
        costs = self._get_api_costs_from_response(model, response)
        return ApiResponse(message=response.content, error=None, costs=costs, files=None)

    async def with_tools_loop(self, prompt: str, file_content_callback: FileRequestCallbackType,
                              follow_up_question_callback: QuestionCallbackType,
                              search_files_callback: SearchFilesCallbackType,
                              create_file_callback: CreateFileCallbackType, model: str, max_tokens: int = 4096,
                              temperature: float = 0.1) -> ApiResponse:
        raise NotImplementedError("deprecated")

    async def agent_loop(self, uuid: str, prompt: str, agent_ids: list[str], model: str, max_tokens: int = 4096,
                         temperature: float = 0.1, max_nr_of_loops: int = 10) -> ApiResponse:
        raise NotImplementedError()

    def estimate_costs(self, model: str, prompt_tokens: int, completion_tokens: int) -> float:
        def costs(nr_tokens, price_per_million: float):
            return (nr_tokens * (price_per_million / (1000. * 1000.))) / 100.  # in cent

        in_price, out_price = MODEL_REGISTRY.get_prices(model)
        return costs(prompt_tokens, in_price) + costs(completion_tokens, out_price)

    async def loop(self, uuid: str, model: str, agent_map: dict[str, Agent], nr_of_loops: int, max_nr_of_loops: int,
                   messages: list, temperature: float, costs: ApiCosts) -> Any:
        raise NotImplementedError()



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Tuple, Generator, Awaitable

import yaml

from menschmachine.llm.types import ApiResponse
from menschmachine.log import get_logger
from menschmachine.platform_adapter import PlatformAdapter
from menschmachine.project import StepCallback, Step, Task


@dataclass
class CodeGenerationStep:
    coroutine: Awaitable[ApiResponse]
    step: Step
    task: Task
    task_number: int
    total_number_of_tasks: int


class AbortGeneration(Exception):
    pass


class FlashSpec(ABC):

    @abstractmethod
    def create_spec(self,
                    uuid: str,
                    model: str,
                    extend: bool = False) -> str:
        pass

    @abstractmethod
    def create_breakdown(self,
                         uuid: str,
                         spec: str,
                         model: str,
                         extend: bool = False) -> str:
        pass

    @abstractmethod
    def create_code(self,
                    uuid: str,
                    spec: str,
                    breakdown: str,
                    model: str,
                    resume_at: int = 0,
                    extend: bool = False,
                    step_completed_callback: StepCallback = StepCallback()) \
            -> Generator[CodeGenerationStep, None, None]:
        pass


def run_check_command(fs: PlatformAdapter) -> Tuple[str, str, int]:
    if os.path.exists(".flashspec.yaml"):
        with open(".flashspec.yaml") as stream:
            try:
                config = yaml.safe_load(stream)
                if "check_command" in config:
                    check_command = config["check_command"]
                    return fs.execute(check_command, user_confirmation=False)
            except yaml.YAMLError as e:
                get_logger().warning("Cannot parse config file", e)
    return "", "", 0



import logging
import os
import subprocess
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Tuple, Callable

from rich.console import Console
from rich.progress import Progress
from ripgrepy import Ripgrepy


class PlatformAdapter(ABC):
    @abstractmethod
    def get_file_list(self, file_filter: Callable[[str], bool] = None) -> list[str]:
        pass

    @abstractmethod
    def read_file(self, file: str) -> str:
        pass

    @abstractmethod
    def write_file(self, file_, content):
        pass

    @abstractmethod
    def ask_user(self, question: str) -> str:
        pass

    @abstractmethod
    def show_progress(self, description: str):
        pass

    @abstractmethod
    def search_files(self, regex: str) -> list[str]:
        pass

    @abstractmethod
    def get_root_dir(self) -> str:
        pass

    @abstractmethod
    def execute(self, cmd: str, user_confirmation: bool = True) -> Tuple[str, str, int]:
        pass


class LocalFilesystemAdapter(PlatformAdapter):

    def get_file_list(self, file_filter: Callable[[str], bool] = None) -> list[str]:
        files = []
        for root, _, filenames in os.walk('.'):
            if '.git' in root.split(os.sep):
                continue
            for filename in filenames:
                f = os.path.join(root, filename)
                if file_filter is None or file_filter(f):
                    files.append(f)
        return files

    def __init__(self, logger: logging.Logger, batch_mode: bool = False):
        self.logger = logger
        self.current_progress_description = None
        self.current_task = None
        self.current_progress = None
        self.batch_mode = batch_mode
        self.console = Console()

    @contextmanager
    def show_progress(self, description: str):
        self.current_progress_description = description
        if self.batch_mode:
            yield
        else:
            with Progress(refresh_per_second=10) as progress:
                self.current_progress = progress
                self.current_task = progress.add_task(description, total=None, transient=True)
                yield self.current_progress
                self.current_progress.update(self.current_task, completed=True)

    def cancel_progress(self):
        if self.current_progress:
            self.current_progress.stop()

    def ask_user(self, question: str) -> str:
        self.cancel_progress()
        a = self.console.input(f"{question}\n> ")
        if self.current_progress_description is not None:
            with self.show_progress(self.current_progress_description):
                return a
        else:
            return a

    def read_file(self, file: str) -> Optional[str]:
        if self.is_readable(file):
            return Path(file).read_text()
        return None

    def write_file(self, file_, content):
        directory = os.path.dirname(file_)
        self.logger.debug(f"Write file {file_}")
        if directory:
            os.makedirs(directory, exist_ok=True)
        with open(file_, 'w') as f:
            f.write(content)

    def execute(self, cmd: str, user_confirmation: bool = True) -> Tuple[str, str, int]:
        self.logger.info(f"Executing cmd {cmd}")
        if not user_confirmation:
            return self.do_execute(cmd)
        while True:
            self.console.print("[bold yellow]Are you sure you want to execute the following command?[/bold yellow]")
            self.console.print(f"[bold cyan]{cmd}[/bold cyan]")
            confirmation = self.ask_user("[bold green]Type 'yes' to confirm or 'no' to cancel:[/bold green]")
            if confirmation.lower() == 'yes':
                return self.do_execute(cmd)
            elif confirmation.lower() == 'no':
                self.logger.info("Command execution cancelled by user.")
                self.console.print("[bold red]Command execution cancelled by user.[/bold red]")
                return "", "Command execution cancelled by user.", 1
            else:
                self.console.print("[bold red]Invalid response. Please answer 'yes' or 'no'.[/bold red]")

    # noinspection PyMethodMayBeStatic
    def do_execute(self, cmd):
        result = subprocess.run(cmd, shell=True, capture_output=True, text=True)
        return result.stdout, result.stderr, result.returncode

    def search_files(self, regex: str) -> list[str]:
        self.logger.info(f"Searching files for {regex}")
        rg = Ripgrepy(regex, self.get_root_dir())
        as_dict = rg.files_with_matches().json().run().as_dict
        matches = list()
        for match in as_dict:
            if match["type"] == "match":
                matches.append(match["data"]["path"]["text"])
        return list(set(matches))

    def get_root_dir(self):
        return os.getcwd()

    @staticmethod
    def is_readable(file: str) -> bool:
        return os.path.isfile(file) and os.access(file, os.R_OK)



from openai import OpenAI

from menschmachine.llm.openrouter import OpenRouterApi
from menschmachine.llm.types import ModelRegistry

_MODEL_REGISTRY = ModelRegistry("gpt-4o-mini")

# Register models with their aliases
_MODEL_REGISTRY.register("gpt-4o-2024-11-20", ["gpt4o", "gpt4", "gpt-4o"], 2.5, 10)
_MODEL_REGISTRY.register("o1-mini", ["gpt4o-mini", "gpt4-mini"], 3, 12)
_MODEL_REGISTRY.register("o1", ["o1", "gpt-01"], 15, 60)
_MODEL_REGISTRY.register("o3-mini", ["o3-mini", "gpt-03-mini", "gpt-03mini"], 1.1, 4.4)
_MODEL_REGISTRY.register("o1-preview", ["o1", "gpt-o1", "o1-preview"], 15, 60)
_MODEL_REGISTRY.register("gpt-4.5-preview", ["gpt45", "gpt-45", "gpt4.5"], 68, 150)
_MODEL_REGISTRY.register("gpt-4o-mini", ["gpt4o-mini", "gpt-4o-mini"], .15, .6)


class OAI(OpenRouterApi):
    def create_client(self):
        return OpenAI()

    def get_model_registry(self) -> ModelRegistry:
        return _MODEL_REGISTRY



import logging
import os
from typing import Callable, Awaitable, Union

from menschmachine.callbacks import Callbacks
from menschmachine.patch import Patch
from menschmachine.platform_adapter import PlatformAdapter


class LocalCallbacks(Callbacks):

    def __init__(self, logger: logging.Logger, fs: PlatformAdapter):
        super().__init__()
        self.fs = fs
        self.logger = logger

    def get_question_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def question_callback(uuid: str, q: dict) -> str:
            question = q['question']
            return self.fs.ask_user(question)

        return question_callback

    def get_file_list_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def filelist_callback(uuid: str, ignore) -> str:
            return str(self.fs.get_file_list())

        return filelist_callback

    def get_run_shell_command_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def run_shell_command_callback(uuid: str, q: dict) -> str:
            cmd = q["command"]
            (stdout, stderr, exit_code) = self.fs.execute(cmd)
            return f"Result of the command is STDOUT: {stdout}\nSTDERR: {stderr}\nEXIT_CODE: {exit_code}"

        return run_shell_command_callback

    def get_search_files_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def search_callback(uuid: str, q: dict) -> str:
            regex = q['regex']
            return str(self.fs.search_files(regex))

        return search_callback

    def get_file_content_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def file_content_callback(uuid: str, q: dict) -> str:
            filepath = q['filepath']
            if os.path.exists(filepath):
                return self.fs.read_file(filepath)
            else:
                return f"FAILED: File {filepath} not found"

        return file_content_callback

    def get_patch_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def patch_callback(uuid: str, q: dict) -> str:
            patch = Patch.from_dirty_xml(q["xml_patch"])
            try:
                for applied in patch.apply_merged(lambda path: self.fs.read_file(path)):
                    self.logger.info(f"Patching {applied.filename}")
                    self.fs.write_file(applied.filename, applied.source)
                return "OK"
            except Exception as e:
                self.logger.error(f"Error applying patch", exc_info=True)
                return f"Patch could not be applied: {e}"

        return patch_callback

    def get_create_file_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:

        def create_file_callback(uuid: str, q: dict) -> str:
            filename_ = q["filepath"]
            if os.path.exists(filename_):
                return f"FAILED: the file {filename_} already exists, you have to use the patch_file tool in this case"
            else:
                self.fs.write_file(filename_, q["content"])
                return "SUCCESS"

        return create_file_callback

    def get_delete_file_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:
        raise NotImplementedError

    def get_create_directory_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:
        raise NotImplementedError

    def get_delete_directory_callback(self) -> Callable[[str, dict], Union[str, Awaitable[str]]]:
        raise NotImplementedError
